from __future__ import annotations
from typing import (Any, TypeVar, Optional, Callable, List, Tuple)
from ...fable_library.array import (fold, filter)
from ...fable_library.date import create
from ...fable_library.int32 import parse
from ...fable_library.option import (value as value_6, map)
from ...fable_library.string import substring
from ...fable_library.util import int32_to_string
from ...fs_hafas_python.context import (Profile, Profile__get_formatStation, Profile__get_departuresGetPasslist, Profile__get_departuresStbFltrEquiv, Profile__get_journeysOutFrwd, Profile__get_transformJourneysQuery)
from ...fs_hafas_python.extensions.date_time_ex import (format_date as format_date_1, format_time as format_time_1)
from ...fs_hafas_python.extra_types import (IndexMap_2, IndexMap_2__get_Item_2B595)
from ...fs_hafas_python.lib.transformations import (Default_LocationsOptions, Coordinate_fromFloat, Default_DeparturesArrivalsOptions, Default_RefreshJourneyOptions, Default_TripsByNameOptions, Default_StopOptions, Default_NearByOptions, Default_ReachableFromOptions, Default_RadarOptions, Default_TripOptions, Default_LinesOptions, Default_ServerOptions, Default_RemarksOptions, Default_JourneysOptions, Default_JourneysFromTripOptions)
from ...fs_hafas_python.types_hafas_client import (ProductType, LocationsOptions, Location, Stop, Station, DeparturesArrivalsOptions, RefreshJourneyOptions, TripsByNameOptions, StopOptions, NearByOptions, ReachableFromOptions, BoundingBox, RadarOptions, TripOptions, LinesOptions, ServerOptions, RemarksOptions, JourneysOptions, StopOver, JourneysFromTripOptions)
from ...fs_hafas_python.types_raw_hafas_client import (JnyFltr, Loc, LocMatchInput, LocMatchRequest, StationBoardRequest, ReconstructionRequest, JourneyMatchRequest, LocDetailsRequest, RawcCrd, RawRing, LocGeoPosRequest, LocGeoReachRequest, RawCrd, RawRect, JourneyGeoPosRequest, JourneyDetailsRequest, LineMatchRequest, ServerInfoRequest, HimSearchRequest, LocViaInput, TripSearchRequest, LocData, SearchOnTripRequest)

_A = TypeVar("_A")

_B = TypeVar("_B")

def ParseIsoString(datetime: str) -> Any:
    year : int = parse(substring(datetime, 0, 4), 511, False, 32) or 0
    month : int = parse(substring(datetime, 5, 2), 511, False, 32) or 0
    day : int = parse(substring(datetime, 8, 2), 511, False, 32) or 0
    hour : int = parse(substring(datetime, 11, 2), 511, False, 32) or 0
    minute : int = parse(substring(datetime, 14, 2), 511, False, 32) or 0
    tz_offset : int = (60 * parse(substring(datetime, 20, 2), 511, False, 32)) or 0
    return create(year, month, day, hour, minute, 0)


def maybe_get_option_value(opt: Optional[_A], getter: Callable[[_A], Optional[_B]]) -> Optional[_B]:
    if opt is None:
        return None
    
    else: 
        return getter(value_6(opt))
    


def get_option_value(opt: Optional[_A], getter: Callable[[_A], Optional[_B]], default_opt: _A) -> _B:
    default_value : _B
    match_value : Optional[_B] = getter(default_opt)
    if match_value is None:
        raise Exception("getOptionValue: value expected")
    
    else: 
        default_value = value_6(match_value)
    
    if opt is None:
        return default_value
    
    else: 
        match_value_1 : Optional[_B] = getter(value_6(opt))
        if match_value_1 is None:
            return default_value
        
        else: 
            return value_6(match_value_1)
        
    


def format_date(dt: Any) -> str:
    return format_date_1(dt, "yyyyMMdd")


def format_time(dt: Any) -> str:
    return format_time_1(dt, "HHmm") + "00"


def format_products_bitmask(profile: Profile, products: IndexMap_2[str, bool]) -> int:
    def folder(bitmask: int, p_1: ProductType, profile: Profile=profile, products: IndexMap_2[str, bool]=products) -> int:
        return p_1.bitmasks[0] | bitmask
    
    def predicate(p: ProductType, profile: Profile=profile, products: IndexMap_2[str, bool]=products) -> bool:
        return IndexMap_2__get_Item_2B595(products, p.id)
    
    return fold(folder, 0, filter(predicate, profile.products))


def make_filters(profile: Profile, products: IndexMap_2[str, bool]) -> List[JnyFltr]:
    bitmask : int = format_products_bitmask(profile, products) or 0
    if bitmask != 0:
        return [JnyFltr("PROD", "INC", int32_to_string(bitmask), None)]
    
    else: 
        return []
    


def location_request(profile: Profile, name: str, opt: Optional[LocationsOptions]=None) -> Tuple[str, LocMatchRequest]:
    def arrow_360(v: LocationsOptions, profile: Profile=profile, name: str=name, opt: Optional[LocationsOptions]=opt) -> Optional[bool]:
        return v.fuzzy
    
    fuzzy : bool = get_option_value(opt, arrow_360, Default_LocationsOptions)
    def arrow_361(v_1: LocationsOptions, profile: Profile=profile, name: str=name, opt: Optional[LocationsOptions]=opt) -> Optional[int]:
        return v_1.results
    
    results : int = get_option_value(opt, arrow_361, Default_LocationsOptions) or 0
    def arrow_362(v_2: LocationsOptions, profile: Profile=profile, name: str=name, opt: Optional[LocationsOptions]=opt) -> Optional[str]:
        return v_2.language
    
    return (get_option_value(opt, arrow_362, Default_LocationsOptions), LocMatchRequest(LocMatchInput(Loc("ALL", name + ("?" if fuzzy else ""), None), results, "S")))


def make_loc_ltype_s(profile: Profile, id: str) -> Loc:
    return Loc("S", None, ("A=1@L=" + Profile__get_formatStation(profile)(id)) + "@")


def make_locl_type_a(location: Location) -> Loc:
    x : int
    match_value : Optional[float] = location.longitude
    if match_value is None:
        raise Exception("location.longitude")
    
    else: 
        x = Coordinate_fromFloat(match_value)
    
    y : int
    match_value_1 : Optional[float] = location.latitude
    if match_value_1 is None:
        raise Exception("location.latitude")
    
    else: 
        y = Coordinate_fromFloat(match_value_1)
    
    xs : str = int32_to_string(x)
    ys : str = int32_to_string(y)
    match_value_2 : Optional[str] = location.address
    if match_value_2 is None:
        return Loc("A", None, ((("A=1@X=" + xs) + "@Y=") + ys) + "@")
    
    else: 
        name : str = match_value_2
        return Loc("A", name, ((((("A=2@O=" + name) + "@X=") + xs) + "@Y=") + ys) + "@")
    


def make_loc_type(profile: Profile, s: Any=None) -> Loc:
    (pattern_matching_result, v_1, v_2) = (None, None, None)
    if str(type(s)) == "\u003cclass \u0027str\u0027\u003e":
        pattern_matching_result = 0
        v_1 = s
    
    elif isinstance(s, Station):
        if s.id is not None:
            pattern_matching_result = 1
            v_2 = s
        
        else: 
            pattern_matching_result = 2
        
    
    else: 
        pattern_matching_result = 2
    
    if pattern_matching_result == 0:
        return make_loc_ltype_s(profile, v_1)
    
    elif pattern_matching_result == 1:
        return make_loc_ltype_s(profile, value_6(v_2.id))
    
    elif pattern_matching_result == 2:
        (pattern_matching_result_1, v_4) = (None, None)
        if isinstance(s, Stop):
            if s.id is not None:
                pattern_matching_result_1 = 0
                v_4 = s
            
            else: 
                pattern_matching_result_1 = 1
            
        
        else: 
            pattern_matching_result_1 = 1
        
        if pattern_matching_result_1 == 0:
            return make_loc_ltype_s(profile, value_6(v_4.id))
        
        elif pattern_matching_result_1 == 1:
            (pattern_matching_result_2, v_6) = (None, None)
            if isinstance(s, Location):
                if s.id is not None:
                    pattern_matching_result_2 = 0
                    v_6 = s
                
                else: 
                    pattern_matching_result_2 = 1
                
            
            else: 
                pattern_matching_result_2 = 1
            
            if pattern_matching_result_2 == 0:
                return make_loc_ltype_s(profile, value_6(v_6.id))
            
            elif pattern_matching_result_2 == 1:
                if isinstance(s, Location):
                    return make_locl_type_a(s)
                
                else: 
                    raise Exception("makeLocType")
                
            
        
    


def station_board_request(profile: Profile, type: str, name: Any=None, opt: Optional[DeparturesArrivalsOptions]=None) -> Tuple[str, StationBoardRequest]:
    def arrow_363(v: DeparturesArrivalsOptions, profile: Profile=profile, type: str=type, name: Any=name, opt: Optional[DeparturesArrivalsOptions]=opt) -> Optional[Any]:
        return v.when
    
    dt : Any = get_option_value(opt, arrow_363, Default_DeparturesArrivalsOptions)
    date : str = format_date(dt)
    time : str = format_time(dt)
    def arrow_364(v_1: DeparturesArrivalsOptions, profile: Profile=profile, type: str=type, name: Any=name, opt: Optional[DeparturesArrivalsOptions]=opt) -> Optional[int]:
        return v_1.duration
    
    duration : int = get_option_value(opt, arrow_364, Default_DeparturesArrivalsOptions) or 0
    def arrow_366(v_2: DeparturesArrivalsOptions, profile: Profile=profile, type: str=type, name: Any=name, opt: Optional[DeparturesArrivalsOptions]=opt) -> Optional[bool]:
        return v_2.stopovers
    
    stopovers : bool = get_option_value(opt, arrow_366, Default_DeparturesArrivalsOptions) if Profile__get_departuresGetPasslist(profile) else False
    def arrow_368(v_3: DeparturesArrivalsOptions, profile: Profile=profile, type: str=type, name: Any=name, opt: Optional[DeparturesArrivalsOptions]=opt) -> Optional[bool]:
        return v_3.include_related_stations
    
    include_related_stations : bool = get_option_value(opt, arrow_368, Default_DeparturesArrivalsOptions) if Profile__get_departuresStbFltrEquiv(profile) else False
    def arrow_369(v_4: DeparturesArrivalsOptions, profile: Profile=profile, type: str=type, name: Any=name, opt: Optional[DeparturesArrivalsOptions]=opt) -> Optional[IndexMap_2[str, bool]]:
        return v_4.products
    
    filters : List[JnyFltr] = make_filters(profile, get_option_value(opt, arrow_369, Default_DeparturesArrivalsOptions))
    def arrow_370(v_5: DeparturesArrivalsOptions, profile: Profile=profile, type: str=type, name: Any=name, opt: Optional[DeparturesArrivalsOptions]=opt) -> Optional[str]:
        return v_5.language
    
    return (get_option_value(opt, arrow_370, Default_DeparturesArrivalsOptions), StationBoardRequest(type, date, time, make_loc_type(profile, name), filters, duration))


def reconstruction_request(profile: Profile, refresh_token: str, opt: Optional[RefreshJourneyOptions]=None) -> Tuple[str, ReconstructionRequest]:
    def arrow_372(v: RefreshJourneyOptions, profile: Profile=profile, refresh_token: str=refresh_token, opt: Optional[RefreshJourneyOptions]=opt) -> Optional[bool]:
        return v.polylines
    
    polylines : bool = get_option_value(opt, arrow_372, Default_RefreshJourneyOptions)
    def arrow_373(v_1: RefreshJourneyOptions, profile: Profile=profile, refresh_token: str=refresh_token, opt: Optional[RefreshJourneyOptions]=opt) -> Optional[bool]:
        return v_1.stopovers
    
    stopovers : bool = get_option_value(opt, arrow_373, Default_RefreshJourneyOptions)
    def arrow_374(v_2: RefreshJourneyOptions, profile: Profile=profile, refresh_token: str=refresh_token, opt: Optional[RefreshJourneyOptions]=opt) -> Optional[bool]:
        return v_2.tickets
    
    tickets : bool = get_option_value(opt, arrow_374, Default_RefreshJourneyOptions)
    def arrow_375(v_3: RefreshJourneyOptions, profile: Profile=profile, refresh_token: str=refresh_token, opt: Optional[RefreshJourneyOptions]=opt) -> Optional[str]:
        return v_3.language
    
    return (get_option_value(opt, arrow_375, Default_RefreshJourneyOptions), ReconstructionRequest(True, stopovers, polylines, tickets, refresh_token))


def journey_match_request(profile: Profile, line_name: str, opt: Optional[TripsByNameOptions]=None) -> Tuple[str, JourneyMatchRequest]:
    def arrow_376(dt: Any, profile: Profile=profile, line_name: str=line_name, opt: Optional[TripsByNameOptions]=opt) -> str:
        return format_date(dt)
    
    def arrow_377(v: TripsByNameOptions, profile: Profile=profile, line_name: str=line_name, opt: Optional[TripsByNameOptions]=opt) -> Optional[Any]:
        return v.when
    
    date : Optional[str] = map(arrow_376, maybe_get_option_value(opt, arrow_377))
    def arrow_378(v_1: TripsByNameOptions, profile: Profile=profile, line_name: str=line_name, opt: Optional[TripsByNameOptions]=opt) -> Optional[str]:
        return "de"
    
    return (get_option_value(opt, arrow_378, Default_TripsByNameOptions), JourneyMatchRequest(line_name, date))


def loc_details_request(profile: Profile, stop: Any=None, opt: Optional[StopOptions]=None) -> Tuple[str, LocDetailsRequest]:
    id : str
    if isinstance(stop, Stop):
        if stop.id is not None:
            id = value_6(stop.id)
        
        else: 
            raise Exception("Stop expected")
        
    
    else: 
        id = stop
    
    def arrow_379(v: StopOptions, profile: Profile=profile, stop: Any=stop, opt: Optional[StopOptions]=opt) -> Optional[str]:
        return v.language
    
    return (get_option_value(opt, arrow_379, Default_StopOptions), LocDetailsRequest([make_loc_ltype_s(profile, id)]))


def loc_geo_pos_request(profile: Profile, location: Location, opt: Optional[NearByOptions]=None) -> Tuple[str, LocGeoPosRequest]:
    def arrow_380(v: NearByOptions, profile: Profile=profile, location: Location=location, opt: Optional[NearByOptions]=opt) -> Optional[int]:
        return v.results
    
    results : int = get_option_value(opt, arrow_380, Default_NearByOptions) or 0
    def arrow_381(v_1: NearByOptions, profile: Profile=profile, location: Location=location, opt: Optional[NearByOptions]=opt) -> Optional[bool]:
        return v_1.stops
    
    stops : bool = get_option_value(opt, arrow_381, Default_NearByOptions)
    def arrow_382(v_2: NearByOptions, profile: Profile=profile, location: Location=location, opt: Optional[NearByOptions]=opt) -> Optional[int]:
        return v_2.distance
    
    distance : int = get_option_value(opt, arrow_382, Default_NearByOptions) or 0
    def arrow_383(v_3: NearByOptions, profile: Profile=profile, location: Location=location, opt: Optional[NearByOptions]=opt) -> Optional[IndexMap_2[str, bool]]:
        return v_3.products
    
    filters : List[JnyFltr] = make_filters(profile, get_option_value(opt, arrow_383, Default_NearByOptions))
    x : int
    match_value : Optional[float] = location.longitude
    if match_value is None:
        raise Exception("location.longitude")
    
    else: 
        x = Coordinate_fromFloat(match_value)
    
    y : int
    match_value_1 : Optional[float] = location.latitude
    if match_value_1 is None:
        raise Exception("location.latitude")
    
    else: 
        y = Coordinate_fromFloat(match_value_1)
    
    def arrow_384(v_4: NearByOptions, profile: Profile=profile, location: Location=location, opt: Optional[NearByOptions]=opt) -> Optional[str]:
        return v_4.language
    
    return (get_option_value(opt, arrow_384, Default_NearByOptions), LocGeoPosRequest(RawRing(RawcCrd(x, y), distance, 0), filters, False, stops, results))


def loc_geo_reach_request(profile: Profile, location: Location, opt: Optional[ReachableFromOptions]=None) -> Tuple[str, LocGeoReachRequest]:
    def arrow_385(v: ReachableFromOptions, profile: Profile=profile, location: Location=location, opt: Optional[ReachableFromOptions]=opt) -> Optional[Any]:
        return v.when
    
    dt : Any = get_option_value(opt, arrow_385, Default_ReachableFromOptions)
    date : str = format_date(dt)
    time : str = format_time(dt)
    def arrow_386(v_1: ReachableFromOptions, profile: Profile=profile, location: Location=location, opt: Optional[ReachableFromOptions]=opt) -> Optional[int]:
        return v_1.max_duration
    
    max_duration : int = get_option_value(opt, arrow_386, Default_ReachableFromOptions) or 0
    def arrow_387(v_2: ReachableFromOptions, profile: Profile=profile, location: Location=location, opt: Optional[ReachableFromOptions]=opt) -> Optional[int]:
        return v_2.max_transfers
    
    max_transfers : int = get_option_value(opt, arrow_387, Default_ReachableFromOptions) or 0
    def arrow_388(v_3: ReachableFromOptions, profile: Profile=profile, location: Location=location, opt: Optional[ReachableFromOptions]=opt) -> Optional[IndexMap_2[str, bool]]:
        return v_3.products
    
    filters : List[JnyFltr] = make_filters(profile, get_option_value(opt, arrow_388, Default_ReachableFromOptions))
    def arrow_389(v_4: ReachableFromOptions, profile: Profile=profile, location: Location=location, opt: Optional[ReachableFromOptions]=opt) -> Optional[str]:
        return "de"
    
    return (get_option_value(opt, arrow_389, Default_ReachableFromOptions), LocGeoReachRequest(make_locl_type_a(location), max_duration, max_transfers, date, time, 120, filters))


def journey_geo_pos_request(profile: Profile, rect: BoundingBox, opt: Optional[RadarOptions]=None) -> Tuple[str, JourneyGeoPosRequest]:
    if rect.north <= rect.south:
        raise Exception("north must be larger than south.")
    
    if rect.east <= rect.west:
        raise Exception("east must be larger than west.")
    
    def arrow_390(v: RadarOptions, profile: Profile=profile, rect: BoundingBox=rect, opt: Optional[RadarOptions]=opt) -> Optional[Any]:
        return v.when
    
    dt : Any = get_option_value(opt, arrow_390, Default_RadarOptions)
    date : str = format_date(dt)
    time : str = format_time(dt)
    def arrow_391(v_1: RadarOptions, profile: Profile=profile, rect: BoundingBox=rect, opt: Optional[RadarOptions]=opt) -> Optional[int]:
        return v_1.results
    
    results : int = get_option_value(opt, arrow_391, Default_RadarOptions) or 0
    def arrow_392(v_2: RadarOptions, profile: Profile=profile, rect: BoundingBox=rect, opt: Optional[RadarOptions]=opt) -> Optional[int]:
        return v_2.duration
    
    duration : int = get_option_value(opt, arrow_392, Default_RadarOptions) or 0
    def arrow_393(v_3: RadarOptions, profile: Profile=profile, rect: BoundingBox=rect, opt: Optional[RadarOptions]=opt) -> Optional[int]:
        return v_3.frames
    
    frames : int = get_option_value(opt, arrow_393, Default_RadarOptions) or 0
    def arrow_394(v_4: RadarOptions, profile: Profile=profile, rect: BoundingBox=rect, opt: Optional[RadarOptions]=opt) -> Optional[IndexMap_2[str, bool]]:
        return v_4.products
    
    filters : List[JnyFltr] = make_filters(profile, get_option_value(opt, arrow_394, Default_RadarOptions))
    def arrow_395(v_5: RadarOptions, profile: Profile=profile, rect: BoundingBox=rect, opt: Optional[RadarOptions]=opt) -> Optional[str]:
        return "de"
    
    return (get_option_value(opt, arrow_395, Default_RadarOptions), JourneyGeoPosRequest(results, False, date, time, RawRect(RawCrd(Coordinate_fromFloat(rect.west), Coordinate_fromFloat(rect.south), None), RawCrd(Coordinate_fromFloat(rect.east), Coordinate_fromFloat(rect.north), None)), duration * 1000, (duration // frames) * 1000, True, filters, "CALC"))


def trip_request(profile: Profile, id: str, name: str, opt: Optional[TripOptions]=None) -> Tuple[str, JourneyDetailsRequest]:
    def arrow_396(v: TripOptions, profile: Profile=profile, id: str=id, name: str=name, opt: Optional[TripOptions]=opt) -> Optional[bool]:
        return v.polyline
    
    polyline : bool = get_option_value(opt, arrow_396, Default_TripOptions)
    def arrow_397(v_1: TripOptions, profile: Profile=profile, id: str=id, name: str=name, opt: Optional[TripOptions]=opt) -> Optional[str]:
        return v_1.language
    
    return (get_option_value(opt, arrow_397, Default_TripOptions), JourneyDetailsRequest(id, name, polyline))


def line_match_request(profile: Profile, query: str, opt: Optional[LinesOptions]=None) -> Tuple[str, LineMatchRequest]:
    def arrow_398(v: LinesOptions, profile: Profile=profile, query: str=query, opt: Optional[LinesOptions]=opt) -> Optional[str]:
        return v.language
    
    return (get_option_value(opt, arrow_398, Default_LinesOptions), LineMatchRequest(query))


def server_info_request(profile: Profile, opt: Optional[ServerOptions]=None) -> Tuple[str, ServerInfoRequest]:
    def arrow_399(v: ServerOptions, profile: Profile=profile, opt: Optional[ServerOptions]=opt) -> Optional[bool]:
        return v.version_info
    
    return ("de", ServerInfoRequest(get_option_value(opt, arrow_399, Default_ServerOptions)))


def him_search_request(profile: Profile, opt: Optional[RemarksOptions]=None) -> Tuple[str, HimSearchRequest]:
    def arrow_400(v: RemarksOptions, profile: Profile=profile, opt: Optional[RemarksOptions]=opt) -> Optional[Any]:
        return v.from_
    
    dt : Any = get_option_value(opt, arrow_400, Default_RemarksOptions)
    date : str = format_date(dt)
    time : str = format_time(dt)
    def arrow_401(v_1: RemarksOptions, profile: Profile=profile, opt: Optional[RemarksOptions]=opt) -> Optional[int]:
        return v_1.results
    
    results : int = get_option_value(opt, arrow_401, Default_RemarksOptions) or 0
    def arrow_402(v_2: RemarksOptions, profile: Profile=profile, opt: Optional[RemarksOptions]=opt) -> Optional[bool]:
        return v_2.polylines
    
    polylines : bool = get_option_value(opt, arrow_402, Default_RemarksOptions)
    def arrow_403(v_3: RemarksOptions, profile: Profile=profile, opt: Optional[RemarksOptions]=opt) -> Optional[IndexMap_2[str, bool]]:
        return v_3.products
    
    filters : List[JnyFltr] = make_filters(profile, get_option_value(opt, arrow_403, Default_RemarksOptions))
    def arrow_404(v_4: RemarksOptions, profile: Profile=profile, opt: Optional[RemarksOptions]=opt) -> Optional[str]:
        return v_4.language
    
    return (get_option_value(opt, arrow_404, Default_RemarksOptions), HimSearchRequest(filters, polylines, results, date, time))


def journey_request(profile: Profile, from_: Any=None, to: Any=None, opt: Optional[JourneysOptions]=None) -> Tuple[str, TripSearchRequest]:
    if opt is None:
        pass
    
    else: 
        opt_1 : JourneysOptions = opt
        if (not Profile__get_journeysOutFrwd(profile)) if (opt_1.arrival is not None) else False:
            raise Exception("opt.arrival is unsupported")
        
        if opt_1.earlier_than is not None:
            raise Exception("opt.earlierThan nyi")
        
        if opt_1.later_than is not None:
            raise Exception("opt.laterThan nyi")
        
    
    def arrow_405(profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Any:
        opt_v_1 : JourneysOptions = opt
        return value_6(opt_v_1.arrival)
    
    def arrow_406(v: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[Any]:
        return v.departure
    
    def arrow_407(v: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[Any]:
        return v.departure
    
    dt : Any = (arrow_405() if (opt.arrival is not None) else get_option_value(opt, arrow_406, Default_JourneysOptions)) if (opt is not None) else get_option_value(opt, arrow_407, Default_JourneysOptions)
    date : str = format_date(dt)
    time : str = format_time(dt)
    out_frwd : bool = False if ((value_6(opt).arrival is not None) if (opt is not None) else False) else True
    def arrow_408(v_1: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[int]:
        return v_1.results
    
    results : int = get_option_value(opt, arrow_408, Default_JourneysOptions) or 0
    def arrow_409(v_2: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[bool]:
        return v_2.stopovers
    
    stopovers : bool = get_option_value(opt, arrow_409, Default_JourneysOptions)
    def arrow_410(v_3: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[int]:
        return v_3.transfers
    
    transfers : int = get_option_value(opt, arrow_410, Default_JourneysOptions) or 0
    def arrow_411(v_4: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[int]:
        return v_4.transfer_time
    
    transfer_time : int = get_option_value(opt, arrow_411, Default_JourneysOptions) or 0
    def arrow_412(v_5: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[bool]:
        return v_5.tickets
    
    tickets : bool = get_option_value(opt, arrow_412, Default_JourneysOptions)
    def arrow_413(v_6: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[bool]:
        return v_6.polylines
    
    polylines : bool = get_option_value(opt, arrow_413, Default_JourneysOptions)
    def arrow_414(v_7: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[IndexMap_2[str, bool]]:
        return v_7.products
    
    filters : List[JnyFltr] = make_filters(profile, get_option_value(opt, arrow_414, Default_JourneysOptions))
    via_loc_l : Optional[List[LocViaInput]]
    def arrow_415(v_8: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[str]:
        return v_8.via
    
    match_value : Optional[str] = maybe_get_option_value(opt, arrow_415)
    via_loc_l = None if (match_value is None) else [LocViaInput(make_loc_ltype_s(profile, match_value))]
    def arrow_416(v_9: JourneysOptions, profile: Profile=profile, from_: Any=from_, to: Any=to, opt: Optional[JourneysOptions]=opt) -> Optional[str]:
        return v_9.language
    
    return (get_option_value(opt, arrow_416, Default_JourneysOptions), Profile__get_transformJourneysQuery(profile)(opt)(TripSearchRequest(stopovers, transfers, transfer_time, [make_loc_type(profile, from_)], via_loc_l, [make_loc_type(profile, to)], filters, [], tickets, True, True, False, polylines, date, time, results, out_frwd, None)))


def search_on_trip_request(profile: Profile, trip_id: str, previous_stopover: StopOver, to: Any=None, opt: Optional[JourneysFromTripOptions]=None) -> Tuple[str, SearchOnTripRequest]:
    prev_stop_id : str
    match_value : Optional[Any] = previous_stopover.stop
    (pattern_matching_result, station_1) = (None, None)
    if match_value is not None:
        if isinstance(value_6(match_value), Station):
            if value_6(match_value).id is not None:
                pattern_matching_result = 0
                station_1 = value_6(match_value)
            
            else: 
                pattern_matching_result = 1
            
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        prev_stop_id = value_6(station_1.id)
    
    elif pattern_matching_result == 1:
        (pattern_matching_result_1, stop_1) = (None, None)
        if match_value is not None:
            if isinstance(value_6(match_value), Stop):
                if value_6(match_value).id is not None:
                    pattern_matching_result_1 = 0
                    stop_1 = value_6(match_value)
                
                else: 
                    pattern_matching_result_1 = 1
                
            
            else: 
                pattern_matching_result_1 = 1
            
        
        else: 
            pattern_matching_result_1 = 1
        
        if pattern_matching_result_1 == 0:
            prev_stop_id = value_6(stop_1.id)
        
        elif pattern_matching_result_1 == 1:
            raise Exception("previousStopover.stop must be a valid stop or station.")
        
    
    dep_at_prev_stop : Any
    match_value_1 : Tuple[Optional[str], Optional[str]] = (previous_stopover.departure, previous_stopover.planned_departure)
    if match_value_1[0] is not None:
        dep_at_prev_stop = ParseIsoString(match_value_1[0])
    
    elif match_value_1[1] is not None:
        dep_at_prev_stop = ParseIsoString(match_value_1[1])
    
    else: 
        raise Exception("previousStopover.(planned)departure is invalid.")
    
    def arrow_417(v: JourneysFromTripOptions, profile: Profile=profile, trip_id: str=trip_id, previous_stopover: StopOver=previous_stopover, to: Any=to, opt: Optional[JourneysFromTripOptions]=opt) -> Optional[bool]:
        return v.tickets
    
    tickets : bool = get_option_value(opt, arrow_417, Default_JourneysFromTripOptions)
    def arrow_418(v_1: JourneysFromTripOptions, profile: Profile=profile, trip_id: str=trip_id, previous_stopover: StopOver=previous_stopover, to: Any=to, opt: Optional[JourneysFromTripOptions]=opt) -> Optional[int]:
        return v_1.transfer_time
    
    transfer_time : int = get_option_value(opt, arrow_418, Default_JourneysFromTripOptions) or 0
    def arrow_419(v_2: JourneysFromTripOptions, profile: Profile=profile, trip_id: str=trip_id, previous_stopover: StopOver=previous_stopover, to: Any=to, opt: Optional[JourneysFromTripOptions]=opt) -> Optional[bool]:
        return v_2.polylines
    
    polylines : bool = get_option_value(opt, arrow_419, Default_JourneysFromTripOptions)
    def arrow_420(v_3: JourneysFromTripOptions, profile: Profile=profile, trip_id: str=trip_id, previous_stopover: StopOver=previous_stopover, to: Any=to, opt: Optional[JourneysFromTripOptions]=opt) -> Optional[bool]:
        return v_3.stopovers
    
    stopovers : bool = get_option_value(opt, arrow_420, Default_JourneysFromTripOptions)
    def arrow_421(v_4: JourneysFromTripOptions, profile: Profile=profile, trip_id: str=trip_id, previous_stopover: StopOver=previous_stopover, to: Any=to, opt: Optional[JourneysFromTripOptions]=opt) -> Optional[IndexMap_2[str, bool]]:
        return v_4.products
    
    filters : List[JnyFltr] = make_filters(profile, get_option_value(opt, arrow_421, Default_JourneysFromTripOptions))
    return ("de", SearchOnTripRequest("JI", trip_id, LocData(make_loc_ltype_s(profile, prev_stop_id), "DEP", format_date(dep_at_prev_stop), format_time(dep_at_prev_stop)), [make_loc_type(profile, to)], filters, stopovers, polylines, transfer_time, tickets))





from __future__ import annotations
from typing import (Optional, List, Any, Tuple, TypeVar,
                    Tuple, Awaitable, Union, Callable, Dict)
import jsonpickle
import json
from functools import reduce
from statistics import mean
from inspect import signature
from .fable_modules.fable_library.array import (map)
from .fable_modules.fs_hafas_python.types_hafas_client import (Profile, IndexMap_2, Station, Stop, Location, ProductType, JourneysOptions, Journeys, RefreshJourneyOptions, Journey, StopOver, JourneysFromTripOptions, TripOptions, Trip, DeparturesArrivalsOptions,
                                                               Alternative, LocationsOptions, StopOptions, Location, NearByOptions, ReachableFromOptions, Duration, BoundingBox, RadarOptions, Movement, TripsByNameOptions, RemarksOptions, Warning, LinesOptions, Line, ServerOptions, ServerInfo)


def stripNone(data):
    if isinstance(data, dict):
        return {k: stripNone(v) for k, v in data.items() if k is not None and v is not None}
    elif isinstance(data, list):
        return [stripNone(item) for item in data if item is not None]
    elif isinstance(data, tuple):
        return tuple(stripNone(item) for item in data if item is not None)
    elif isinstance(data, set):
        return {stripNone(item) for item in data if item is not None}
    else:
        if hasattr(data, '__dict__'):
            return stripNone(data.__dict__.copy())
        else:
            return data


def simplify(value: Any) -> Any:
    """
    simplify to JSON-able python dicts, strip null values
    """
    return json.loads(json_encode(value, stripNones=True))


def json_encode(value: Any, stripNones: Optional[bool] = True) -> str:
    """
    Return a JSON formatted representation of value, a Python object
    """
    remapped = stripNone(value) if stripNones else value
    return jsonpickle.encode(remapped, unpicklable=False)


def _replace(o: Dict, replacement: Tuple[str, Any]):
    """
    exec replacement in o
    replacement is Tuple[key,replacable]
    replacable may be key, list of keys or callable
    """
    if isinstance(o, Dict):
        key = replacement[0]
        replacable = replacement[1]
        if isinstance(replacable, str):
            o[key] = o[replacable]
        elif isinstance(replacable, List):
            o[key] = reduce(lambda acc, v: acc[v], replacable, o)
        elif isinstance(replacable, Callable):
            sig = signature(replacable)
            if len(sig.parameters) == 0:
                o[key] = replacable()
            elif len(sig.parameters) == 1:
                o[key] = replacable(o)


def replace(o: Union[Dict, List[Dict]], replacements: List[Tuple[str, Any]]) -> Any:
    """
    exec replacements in o, return o
    """
    if isinstance(o, List):
        for d in o:
            for r in replacements:
                _replace(d, r)
    elif isinstance(o, Dict):
        for r in replacements:
            _replace(o, r)

    return o


def to_location(s: Union[Station, Stop, Location]) -> Location:
    if isinstance(s, Location):
        return s

    elif isinstance(s, Stop):
        location: Location = s.location
        location.id = s.id
        location.name = s.name
        return location

    elif isinstance(s, Station):
        location: Location = s.location
        location.id = s.id
        location.name = s.name
        return location

    else:
        return None


def to_locations(sx: List[Union[Station, Stop, Location]]) -> List[Location]:
    """
    convert list of Station | Stop | Location to List of Location
    """
    # workaround: error in type of map, should accept Optional
    return map(to_location, sx, None)  # pytype: disable=wrong-arg-types


def to_name(s: Union[Station, Stop, Location, Line]) -> str:
    """
    get name of Station | Stop | Location | Line
    """
    if isinstance(s, Location):
        return s.name

    elif isinstance(s, Stop):
        return s.name

    elif isinstance(s, Station):
        return s.name

    elif isinstance(s, Line):
        return s.name

    else:
        return ""


def to_bounding_box(sx1: List[Union[Station, Stop, Location]], sx2: List[Union[Station, Stop, Location]]) -> BoundingBox:
    locs_1 = to_locations(sx1)
    locs_2 = to_locations(sx2)
    if len(locs_1) > 0 and len(locs_2) > 0:
        n = max(locs_1[0].latitude, locs_2[0].latitude)
        s = min(locs_1[0].latitude, locs_2[0].latitude)
        w = min(locs_1[0].longitude, locs_2[0].longitude)
        e = max(locs_1[0].longitude, locs_2[0].longitude)
        return BoundingBox(n, w, s, e)
    else:
        return BoundingBox(52.2735877, 8.0596000,  51.7128598, 8.7404385)


def centroid(movements: list[Movement], rect: BoundingBox):
    if len(movements) > 0:
        _x = mean([m.location.latitude for m in movements])
        _y = mean([m.location.longitude for m in movements])
        return(_x, _y)
    else:
        _x = mean([rect.north, rect.south])
        _y = mean([rect.west, rect.east])
        return(_x, _y)



from __future__ import annotations
from typing import (TypeVar, Any, Callable, Generic, Optional, Tuple, List)
from .array import (singleton as singleton_1, try_find_back as try_find_back_1, try_find_index_back as try_find_index_back_1, fold_back as fold_back_1, fold_back2 as fold_back2_1, try_head as try_head_1, try_item as try_item_1, map_fold as map_fold_1, map_fold_back as map_fold_back_1, reduce_back as reduce_back_1, reverse as reverse_1, scan_back as scan_back_1, pairwise as pairwise_1, map as map_1, split_into as split_into_1, windowed as windowed_1, transpose as transpose_1, permute as permute_1, chunk_by_size as chunk_by_size_1)
from .fsharp_core import (Operators_NullArg, Operators_Lock)
from .global_ import (IGenericAdder_1, IGenericAverager_1, SR_indexOutOfBounds)
from .list import (FSharpList, to_array as to_array_1, of_array as of_array_1, of_seq as of_seq_1, try_head as try_head_2, is_empty as is_empty_1, try_item as try_item_2, length as length_1)
from .option import (value as value_1, some)
from .reflection import (TypeInfo, class_type)
from .types import to_string
from .util import (get_enumerator, IEnumerator, to_iterator, IDisposable, is_disposable, dispose as dispose_2, IEnumerable, is_array_like, IEqualityComparer, equals, ignore, clear, IComparer)

_A_ = TypeVar("_A_")

_T = TypeVar("_T")

_U = TypeVar("_U")

_STATE = TypeVar("_STATE")

_B_ = TypeVar("_B_")

_T1 = TypeVar("_T1")

_T2 = TypeVar("_T2")

_T3 = TypeVar("_T3")

_RESULT = TypeVar("_RESULT")

SR_enumerationAlreadyFinished : str = "Enumeration already finished."

SR_enumerationNotStarted : str = "Enumeration has not started. Call MoveNext."

SR_inputSequenceEmpty : str = "The input sequence was empty."

SR_inputSequenceTooLong : str = "The input sequence contains more than one element."

SR_keyNotFoundAlt : str = "An index satisfying the predicate was not found in the collection."

SR_notEnoughElements : str = "The input sequence has an insufficient number of elements."

SR_resetNotSupported : str = "Reset is not supported on this enumerator."

def Enumerator_noReset() -> _A_:
    raise Exception(SR_resetNotSupported)


def Enumerator_notStarted() -> _A_:
    raise Exception(SR_enumerationNotStarted)


def Enumerator_alreadyFinished() -> _A_:
    raise Exception(SR_enumerationAlreadyFinished)


def expr_147(gen0: TypeInfo) -> TypeInfo:
    return class_type("SeqModule.Enumerator.Seq", [gen0], Enumerator_Seq)


class Enumerator_Seq(Generic[_T]):
    def __init__(self, f: Callable[[], IEnumerator[_T]]) -> None:
        self.f = f
    
    def __str__(self) -> str:
        xs : Enumerator_Seq[_T] = self
        max_count : int = 4
        i : int = 0
        str_1 : str = "seq ["
        with get_enumerator(xs) as e:
            while e.System_Collections_IEnumerator_MoveNext() if (i < max_count) else False:
                if i > 0:
                    str_1 = str_1 + "; "
                
                str_1 = str_1 + to_string(e.System_Collections_Generic_IEnumerator_00601_get_Current())
                i = (i + 1) or 0
            if i == max_count:
                str_1 = str_1 + "; ..."
            
            return str_1 + "]"
    
    def GetEnumerator(self) -> IEnumerator[_T]:
        x : Enumerator_Seq[_T] = self
        return x.f()
    
    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())
    
    def System_Collections_IEnumerable_GetEnumerator(self) -> IEnumerator[Any]:
        x : Enumerator_Seq[_T] = self
        return x.f()
    

Enumerator_Seq_reflection = expr_147

def Enumerator_Seq__ctor_673A07F2(f: Callable[[], IEnumerator[_T]]) -> Enumerator_Seq[_T]:
    return Enumerator_Seq(f)


def expr_148(gen0: TypeInfo) -> TypeInfo:
    return class_type("SeqModule.Enumerator.FromFunctions`1", [gen0], Enumerator_FromFunctions_1)


class Enumerator_FromFunctions_1(IDisposable, Generic[_T]):
    def __init__(self, current: Callable[[], _T], next_1: Callable[[], bool], dispose: Callable[[], None]) -> None:
        self.current = current
        self.next = next_1
        self.dispose = dispose
    
    def System_Collections_Generic_IEnumerator_00601_get_Current(self) -> _T:
        __ : Enumerator_FromFunctions_1[_T] = self
        return __.current()
    
    def System_Collections_IEnumerator_get_Current(self) -> Any:
        __ : Enumerator_FromFunctions_1[_T] = self
        return __.current()
    
    def System_Collections_IEnumerator_MoveNext(self) -> bool:
        __ : Enumerator_FromFunctions_1[_T] = self
        return __.next()
    
    def System_Collections_IEnumerator_Reset(self) -> None:
        Enumerator_noReset()
    
    def Dispose(self) -> None:
        __ : Enumerator_FromFunctions_1[_T] = self
        __.dispose()
    

Enumerator_FromFunctions_1_reflection = expr_148

def Enumerator_FromFunctions_1__ctor_58C54629(current: Callable[[], _T], next_1: Callable[[], bool], dispose: Callable[[], None]) -> Enumerator_FromFunctions_1[_T]:
    return Enumerator_FromFunctions_1(current, next_1, dispose)


def Enumerator_cast(e: IEnumerator[Any]) -> IEnumerator[_T]:
    def current(e: IEnumerator[Any]=e) -> _T:
        return e.System_Collections_IEnumerator_get_Current()
    
    def next_1(e: IEnumerator[Any]=e) -> bool:
        return e.System_Collections_IEnumerator_MoveNext()
    
    def dispose(e: IEnumerator[Any]=e) -> None:
        if is_disposable(e):
            dispose_2(e)
        
    
    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def Enumerator_concat(sources: IEnumerable[_U]) -> IEnumerator[_T]:
    outer_opt : Optional[IEnumerator[_U]] = None
    inner_opt : Optional[IEnumerator[_T]] = None
    started : bool = False
    finished : bool = False
    curr : Optional[_T] = None
    def finish(sources: IEnumerable[_U]=sources) -> None:
        nonlocal finished, inner_opt, outer_opt
        finished = True
        if inner_opt is not None:
            inner : IEnumerator[_T] = inner_opt
            try: 
                dispose_2(inner)
            
            finally: 
                inner_opt = None
            
        
        if outer_opt is not None:
            outer : IEnumerator[_U] = outer_opt
            try: 
                dispose_2(outer)
            
            finally: 
                outer_opt = None
            
        
    
    def current(sources: IEnumerable[_U]=sources) -> _T:
        if not started:
            Enumerator_notStarted()
        
        elif finished:
            Enumerator_alreadyFinished()
        
        if curr is not None:
            return value_1(curr)
        
        else: 
            return Enumerator_alreadyFinished()
        
    
    def next_1(sources: IEnumerable[_U]=sources) -> bool:
        nonlocal started
        if not started:
            started = True
        
        if finished:
            return False
        
        else: 
            res : Optional[bool] = None
            while res is None:
                nonlocal curr, inner_opt, outer_opt
                match_value : Tuple[Optional[IEnumerator[_U]], Optional[IEnumerator[_T]]] = (outer_opt, inner_opt)
                if match_value[0] is not None:
                    if match_value[1] is not None:
                        inner_1 : IEnumerator[_T] = match_value[1]
                        if inner_1.System_Collections_IEnumerator_MoveNext():
                            curr = some(inner_1.System_Collections_Generic_IEnumerator_00601_get_Current())
                            res = True
                        
                        else: 
                            try: 
                                dispose_2(inner_1)
                            
                            finally: 
                                inner_opt = None
                            
                        
                    
                    else: 
                        outer_1 : IEnumerator[_U] = match_value[0]
                        if outer_1.System_Collections_IEnumerator_MoveNext():
                            ie : _U = outer_1.System_Collections_Generic_IEnumerator_00601_get_Current()
                            def arrow_149(__unit: Any=None) -> IEnumerator[_T]:
                                copy_of_struct : _U = ie
                                return get_enumerator(copy_of_struct)
                            
                            inner_opt = arrow_149()
                        
                        else: 
                            finish()
                            res = False
                        
                    
                
                else: 
                    outer_opt = get_enumerator(sources)
                
            return value_1(res)
        
    
    def dispose(sources: IEnumerable[_U]=sources) -> None:
        if not finished:
            finish()
        
    
    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def Enumerator_enumerateThenFinally(f: Callable[[], None], e: IEnumerator[_T]) -> IEnumerator[_T]:
    def current(f: Callable[[], None]=f, e: IEnumerator[_T]=e) -> _T:
        return e.System_Collections_Generic_IEnumerator_00601_get_Current()
    
    def next_1(f: Callable[[], None]=f, e: IEnumerator[_T]=e) -> bool:
        return e.System_Collections_IEnumerator_MoveNext()
    
    def dispose(f: Callable[[], None]=f, e: IEnumerator[_T]=e) -> None:
        try: 
            dispose_2(e)
        
        finally: 
            f()
        
    
    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def Enumerator_generateWhileSome(openf: Callable[[], _T], compute: Callable[[_T], Optional[_U]], closef: Callable[[_T], None]) -> IEnumerator[_U]:
    started : bool = False
    curr : Optional[_U] = None
    state : Optional[_T] = some(openf())
    def dispose(openf: Callable[[], _T]=openf, compute: Callable[[_T], Optional[_U]]=compute, closef: Callable[[_T], None]=closef) -> None:
        nonlocal state
        if state is not None:
            x_1 : _T = value_1(state)
            try: 
                closef(x_1)
            
            finally: 
                state = None
            
        
    
    def finish(openf: Callable[[], _T]=openf, compute: Callable[[_T], Optional[_U]]=compute, closef: Callable[[_T], None]=closef) -> None:
        nonlocal curr
        try: 
            dispose()
        
        finally: 
            curr = None
        
    
    def current(openf: Callable[[], _T]=openf, compute: Callable[[_T], Optional[_U]]=compute, closef: Callable[[_T], None]=closef) -> _U:
        if not started:
            Enumerator_notStarted()
        
        if curr is not None:
            return value_1(curr)
        
        else: 
            return Enumerator_alreadyFinished()
        
    
    def next_1(openf: Callable[[], _T]=openf, compute: Callable[[_T], Optional[_U]]=compute, closef: Callable[[_T], None]=closef) -> bool:
        nonlocal started, curr
        if not started:
            started = True
        
        if state is not None:
            s : _T = value_1(state)
            match_value_1 : Optional[_U]
            try: 
                match_value_1 = compute(s)
            
            except Exception as match_value:
                finish()
                raise match_value
            
            if match_value_1 is not None:
                curr = match_value_1
                return True
            
            else: 
                finish()
                return False
            
        
        else: 
            return False
        
    
    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def Enumerator_unfold(f: Callable[[_STATE], Optional[Tuple[_T, _STATE]]], state: _STATE) -> IEnumerator[_T]:
    curr : Optional[Tuple[_T, _STATE]] = None
    acc : _STATE = state
    def current(f: Callable[[_STATE], Optional[Tuple[_T, _STATE]]]=f, state: _STATE=state) -> _T:
        if curr is not None:
            x : _T = curr[0]
            st : _STATE = curr[1]
            return x
        
        else: 
            return Enumerator_notStarted()
        
    
    def next_1(f: Callable[[_STATE], Optional[Tuple[_T, _STATE]]]=f, state: _STATE=state) -> bool:
        nonlocal curr, acc
        curr = f(acc)
        if curr is not None:
            x_1 : _T = curr[0]
            st_1 : _STATE = curr[1]
            acc = st_1
            return True
        
        else: 
            return False
        
    
    def dispose(f: Callable[[_STATE], Optional[Tuple[_T, _STATE]]]=f, state: _STATE=state) -> None:
        pass
    
    return Enumerator_FromFunctions_1__ctor_58C54629(current, next_1, dispose)


def index_not_found() -> _A_:
    raise Exception(SR_keyNotFoundAlt)


def check_non_null(arg_name: str, arg: _A_) -> None:
    if arg is None:
        Operators_NullArg(arg_name)
    


def mk_seq(f: Callable[[], IEnumerator[_T]]) -> IEnumerable[_T]:
    return Enumerator_Seq__ctor_673A07F2(f)


def of_seq(xs: IEnumerable[_T]) -> IEnumerator[_T]:
    check_non_null("source", xs)
    return get_enumerator(xs)


def delay(generator: Callable[[], IEnumerable[_T]]) -> IEnumerable[_T]:
    def arrow_150(generator: Callable[[], IEnumerable[_T]]=generator) -> IEnumerator[_T]:
        return get_enumerator(generator())
    
    return mk_seq(arrow_150)


def concat(sources: IEnumerable[_A_]) -> IEnumerable[_T]:
    def arrow_151(sources: IEnumerable[_A_]=sources) -> IEnumerator[_T]:
        return Enumerator_concat(sources)
    
    return mk_seq(arrow_151)


def unfold(generator: Callable[[_STATE], Optional[Tuple[_T, _STATE]]], state: _STATE) -> IEnumerable[_T]:
    def arrow_152(generator: Callable[[_STATE], Optional[Tuple[_T, _STATE]]]=generator, state: _STATE=state) -> IEnumerator[_T]:
        return Enumerator_unfold(generator, state)
    
    return mk_seq(arrow_152)


def empty() -> IEnumerable[_T]:
    def arrow_153(__unit: Any=None) -> IEnumerable[_T]:
        return []
    
    return delay(arrow_153)


def singleton(x: Optional[_T]=None) -> IEnumerable[_T]:
    def arrow_154(x: _T=x) -> IEnumerable[_T]:
        return singleton_1(x, None)
    
    return delay(arrow_154)


def of_array(arr: List[_T]) -> IEnumerable[_T]:
    return arr


def to_array(xs: IEnumerable[_T]) -> List[_T]:
    if isinstance(xs, FSharpList):
        return to_array_1(xs)
    
    else: 
        return list(xs)
    


def of_list(xs: FSharpList[_T]) -> IEnumerable[_T]:
    return xs


def to_list(xs: IEnumerable[_T]) -> IEnumerable[_T]:
    if is_array_like(xs):
        return of_array_1(xs)
    
    elif isinstance(xs, FSharpList):
        return xs
    
    else: 
        return of_seq_1(xs)
    


def generate(create: Callable[[], _A_], compute: Callable[[_A_], Optional[_B_]], dispose: Callable[[_A_], None]) -> IEnumerable[_B_]:
    def arrow_155(create: Callable[[], _A_]=create, compute: Callable[[_A_], Optional[_B_]]=compute, dispose: Callable[[_A_], None]=dispose) -> IEnumerator[_B_]:
        return Enumerator_generateWhileSome(create, compute, dispose)
    
    return mk_seq(arrow_155)


def generate_indexed(create: Callable[[], _A_], compute: Callable[[int, _A_], Optional[_B_]], dispose: Callable[[_A_], None]) -> IEnumerable[_B_]:
    def arrow_157(create: Callable[[], _A_]=create, compute: Callable[[int, _A_], Optional[_B_]]=compute, dispose: Callable[[_A_], None]=dispose) -> IEnumerator[_B_]:
        i : int = -1
        def arrow_156(x: Optional[_A_]=None) -> Optional[_B_]:
            nonlocal i
            i = (i + 1) or 0
            return compute(i, x)
        
        return Enumerator_generateWhileSome(create, arrow_156, dispose)
    
    return mk_seq(arrow_157)


def append(xs: IEnumerable[_T], ys: IEnumerable[_T]) -> IEnumerable[_T]:
    return concat([xs, ys])


def cast(xs: IEnumerable[Any]) -> IEnumerable[_A_]:
    def arrow_158(xs: IEnumerable[Any]=xs) -> IEnumerator[_A_]:
        check_non_null("source", xs)
        return Enumerator_cast(get_enumerator(xs))
    
    return mk_seq(arrow_158)


def choose(chooser: Callable[[_T], Optional[_U]], xs: IEnumerable[_T]) -> IEnumerable[_U]:
    def arrow_159(chooser: Callable[[_T], Optional[_U]]=chooser, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_160(e: IEnumerator[_T], chooser: Callable[[_T], Optional[_U]]=chooser, xs: IEnumerable[_T]=xs) -> Optional[_U]:
        curr : Optional[_U] = None
        while e.System_Collections_IEnumerator_MoveNext() if (curr is None) else False:
            curr = chooser(e.System_Collections_Generic_IEnumerator_00601_get_Current())
        return curr
    
    def arrow_161(e_1: IEnumerator[_T], chooser: Callable[[_T], Optional[_U]]=chooser, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate(arrow_159, arrow_160, arrow_161)


def compare_with(comparer: Callable[[_T, _T], int], xs: IEnumerable[_T], ys: IEnumerable[_T]) -> int:
    with of_seq(xs) as e1:
        with of_seq(ys) as e2:
            c : int = 0
            b1 : bool = e1.System_Collections_IEnumerator_MoveNext()
            b2 : bool = e2.System_Collections_IEnumerator_MoveNext()
            while b2 if (b1 if (c == 0) else False) else False:
                c = comparer(e1.System_Collections_Generic_IEnumerator_00601_get_Current(), e2.System_Collections_Generic_IEnumerator_00601_get_Current()) or 0
                if c == 0:
                    b1 = e1.System_Collections_IEnumerator_MoveNext()
                    b2 = e2.System_Collections_IEnumerator_MoveNext()
                
            if c != 0:
                return c
            
            elif b1:
                return 1
            
            elif b2:
                return -1
            
            else: 
                return 0
            


def contains(value: _T, xs: IEnumerable[_T], comparer: IEqualityComparer[Any]) -> bool:
    with of_seq(xs) as e:
        found : bool = False
        while e.System_Collections_IEnumerator_MoveNext() if (not found) else False:
            found = comparer.Equals(value, e.System_Collections_Generic_IEnumerator_00601_get_Current())
        return found


def enumerate_from_functions(create: Callable[[], _A_], move_next: Callable[[_A_], bool], current: Callable[[_A_], _B_]) -> IEnumerable[_B_]:
    def arrow_162(x: Optional[_A_]=None, create: Callable[[], _A_]=create, move_next: Callable[[_A_], bool]=move_next, current: Callable[[_A_], _B_]=current) -> Optional[_B_]:
        return some(current(x)) if move_next(x) else None
    
    def arrow_163(x_1: Optional[_A_]=None, create: Callable[[], _A_]=create, move_next: Callable[[_A_], bool]=move_next, current: Callable[[_A_], _B_]=current) -> None:
        match_value : Any = x_1
        if is_disposable(match_value):
            dispose_2(match_value)
        
    
    return generate(create, arrow_162, arrow_163)


def enumerate_then_finally(source: IEnumerable[_T], compensation: Callable[[], None]) -> IEnumerable[_T]:
    compensation_1 : Callable[[], None] = compensation
    def arrow_164(source: IEnumerable[_T]=source, compensation: Callable[[], None]=compensation) -> IEnumerator[_T]:
        try: 
            return Enumerator_enumerateThenFinally(compensation_1, of_seq(source))
        
        except Exception as match_value:
            compensation_1()
            raise match_value
        
    
    return mk_seq(arrow_164)


def enumerate_using(resource: _T, source: Callable[[_T], _A_]) -> IEnumerable[_U]:
    def compensation(resource: _T=resource, source: Callable[[_T], _A_]=source) -> None:
        if equals(resource, None):
            pass
        
        else: 
            copy_of_struct : _T = resource
            dispose_2(copy_of_struct)
        
    
    def arrow_166(resource: _T=resource, source: Callable[[_T], _A_]=source) -> IEnumerator[_T]:
        try: 
            def arrow_165(xs: IEnumerable[_T]) -> IEnumerator[_T]:
                return of_seq(xs)
            
            return Enumerator_enumerateThenFinally(compensation, arrow_165(source(resource)))
        
        except Exception as match_value_1:
            compensation()
            raise match_value_1
        
    
    return mk_seq(arrow_166)


def enumerate_while(guard: Callable[[], bool], xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_167(i: int, guard: Callable[[], bool]=guard, xs: IEnumerable[_T]=xs) -> Optional[Tuple[IEnumerable[_T], int]]:
        return ((xs, i + 1)) if guard() else None
    
    return concat(unfold(arrow_167, 0))


def filter(f: Callable[[_T], bool], xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def chooser(x: Optional[_T]=None, f: Callable[[_T], bool]=f, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        if f(x):
            return some(x)
        
        else: 
            return None
        
    
    return choose(chooser, xs)


def exists(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> bool:
    with of_seq(xs) as e:
        found : bool = False
        while e.System_Collections_IEnumerator_MoveNext() if (not found) else False:
            found = predicate(e.System_Collections_Generic_IEnumerator_00601_get_Current())
        return found


def exists2(predicate: Callable[[_T1, _T2], bool], xs: IEnumerable[_T1], ys: IEnumerable[_T2]) -> bool:
    with of_seq(xs) as e1:
        with of_seq(ys) as e2:
            found : bool = False
            while e2.System_Collections_IEnumerator_MoveNext() if (e1.System_Collections_IEnumerator_MoveNext() if (not found) else False) else False:
                found = predicate(e1.System_Collections_Generic_IEnumerator_00601_get_Current(), e2.System_Collections_Generic_IEnumerator_00601_get_Current())
            return found


def exactly_one(xs: IEnumerable[_T]) -> _T:
    with of_seq(xs) as e:
        if e.System_Collections_IEnumerator_MoveNext():
            v : _T = e.System_Collections_Generic_IEnumerator_00601_get_Current()
            if e.System_Collections_IEnumerator_MoveNext():
                raise Exception((SR_inputSequenceTooLong + "\\nParameter name: ") + "source")
            
            else: 
                return v
            
        
        else: 
            raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "source")
        


def try_exactly_one(xs: IEnumerable[_T]) -> Optional[_T]:
    with of_seq(xs) as e:
        if e.System_Collections_IEnumerator_MoveNext():
            v : _T = e.System_Collections_Generic_IEnumerator_00601_get_Current()
            if e.System_Collections_IEnumerator_MoveNext():
                return None
            
            else: 
                return some(v)
            
        
        else: 
            return None
        


def try_find(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> Optional[_T]:
    with of_seq(xs) as e:
        res : Optional[_T] = None
        while e.System_Collections_IEnumerator_MoveNext() if (res is None) else False:
            c : _T = e.System_Collections_Generic_IEnumerator_00601_get_Current()
            if predicate(c):
                res = some(c)
            
        return res


def find(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> _T:
    match_value : Optional[_T] = try_find(predicate, xs)
    if match_value is None:
        return index_not_found()
    
    else: 
        return value_1(match_value)
    


def try_find_back(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> Optional[_T]:
    return try_find_back_1(predicate, to_array(xs))


def find_back(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> _T:
    match_value : Optional[_T] = try_find_back(predicate, xs)
    if match_value is None:
        return index_not_found()
    
    else: 
        return value_1(match_value)
    


def try_find_index(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> Optional[int]:
    with of_seq(xs) as e:
        def loop(i_mut: int, predicate: Callable[[_T], bool]=predicate, xs: IEnumerable[_T]=xs) -> Optional[int]:
            while True:
                (i,) = (i_mut,)
                if e.System_Collections_IEnumerator_MoveNext():
                    if predicate(e.System_Collections_Generic_IEnumerator_00601_get_Current()):
                        return i
                    
                    else: 
                        i_mut = i + 1
                        continue
                    
                
                else: 
                    return None
                
                break
        
        loop : Callable[[int], Optional[int]] = loop
        return loop(0)


def find_index(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> int:
    match_value : Optional[int] = try_find_index(predicate, xs)
    if match_value is None:
        return index_not_found()
    
    else: 
        return match_value
    


def try_find_index_back(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> Optional[int]:
    return try_find_index_back_1(predicate, to_array(xs))


def find_index_back(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> int:
    match_value : Optional[int] = try_find_index_back(predicate, xs)
    if match_value is None:
        return index_not_found()
    
    else: 
        return match_value
    


def fold(folder: Callable[[_STATE, _T], _STATE], state: _STATE, xs: IEnumerable[_T]) -> _STATE:
    with of_seq(xs) as e:
        acc : _STATE = state
        while e.System_Collections_IEnumerator_MoveNext():
            acc = folder(acc, e.System_Collections_Generic_IEnumerator_00601_get_Current())
        return acc


def fold_back(folder: Callable[[_T, _A_], _A_], xs: IEnumerable[_T], state: _A_) -> _A_:
    return fold_back_1(folder, to_array(xs), state)


def fold2(folder: Callable[[_STATE, _T1, _T2], _STATE], state: _STATE, xs: IEnumerable[_T1], ys: IEnumerable[_T2]) -> _STATE:
    with of_seq(xs) as e1:
        with of_seq(ys) as e2:
            acc : _STATE = state
            while e2.System_Collections_IEnumerator_MoveNext() if e1.System_Collections_IEnumerator_MoveNext() else False:
                acc = folder(acc, e1.System_Collections_Generic_IEnumerator_00601_get_Current(), e2.System_Collections_Generic_IEnumerator_00601_get_Current())
            return acc


def fold_back2(folder: Callable[[_T1, _T2, _STATE], _STATE], xs: IEnumerable[_T1], ys: IEnumerable[_T2], state: _STATE) -> _STATE:
    return fold_back2_1(folder, to_array(xs), to_array(ys), state)


def for_all(predicate: Callable[[_A_], bool], xs: IEnumerable[_A_]) -> bool:
    def arrow_168(x: Optional[_A_]=None, predicate: Callable[[_A_], bool]=predicate, xs: IEnumerable[_A_]=xs) -> bool:
        return not predicate(x)
    
    return not exists(arrow_168, xs)


def for_all2(predicate: Callable[[_A_, _B_], bool], xs: IEnumerable[_A_], ys: IEnumerable[_B_]) -> bool:
    def arrow_169(x: _A_, y: _B_, predicate: Callable[[_A_, _B_], bool]=predicate, xs: IEnumerable[_A_]=xs, ys: IEnumerable[_B_]=ys) -> bool:
        return not predicate(x, y)
    
    return not exists2(arrow_169, xs, ys)


def try_head(xs: IEnumerable[_T]) -> Optional[_T]:
    if is_array_like(xs):
        return try_head_1(xs)
    
    elif isinstance(xs, FSharpList):
        return try_head_2(xs)
    
    else: 
        with of_seq(xs) as e:
            if e.System_Collections_IEnumerator_MoveNext():
                return some(e.System_Collections_Generic_IEnumerator_00601_get_Current())
            
            else: 
                return None
            
    


def head(xs: IEnumerable[_T]) -> _T:
    match_value : Optional[_T] = try_head(xs)
    if match_value is None:
        raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "source")
    
    else: 
        return value_1(match_value)
    


def initialize(count: int, f: Callable[[int], _A_]) -> IEnumerable[_A_]:
    def arrow_170(i: int, count: int=count, f: Callable[[int], _A_]=f) -> Optional[Tuple[_A_, int]]:
        return ((f(i), i + 1)) if (i < count) else None
    
    return unfold(arrow_170, 0)


def initialize_infinite(f: Callable[[int], _A_]) -> IEnumerable[_A_]:
    return initialize(2147483647, f)


def is_empty(xs: IEnumerable[_T]) -> bool:
    if is_array_like(xs):
        return len(xs) == 0
    
    elif isinstance(xs, FSharpList):
        return is_empty_1(xs)
    
    else: 
        with of_seq(xs) as e:
            return not e.System_Collections_IEnumerator_MoveNext()
    


def try_item(index: int, xs: IEnumerable[_T]) -> Optional[_T]:
    if is_array_like(xs):
        return try_item_1(index, xs)
    
    elif isinstance(xs, FSharpList):
        return try_item_2(index, xs)
    
    else: 
        with of_seq(xs) as e:
            def loop(index_1_mut: int, index: int=index, xs: IEnumerable[_T]=xs) -> Optional[_T]:
                while True:
                    (index_1,) = (index_1_mut,)
                    if not e.System_Collections_IEnumerator_MoveNext():
                        return None
                    
                    elif index_1 == 0:
                        return some(e.System_Collections_Generic_IEnumerator_00601_get_Current())
                    
                    else: 
                        index_1_mut = index_1 - 1
                        continue
                    
                    break
            
            loop : Callable[[int], Optional[_T]] = loop
            return loop(index)
    


def item(index: int, xs: IEnumerable[_T]) -> _T:
    match_value : Optional[_T] = try_item(index, xs)
    if match_value is None:
        raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "index")
    
    else: 
        return value_1(match_value)
    


def iterate(action: Callable[[_A_], None], xs: IEnumerable[_A_]) -> None:
    def arrow_171(unit_var0: None, x: _A_, action: Callable[[_A_], None]=action, xs: IEnumerable[_A_]=xs) -> None:
        action(x)
    
    fold(arrow_171, None, xs)


def iterate2(action: Callable[[_A_, _B_], None], xs: IEnumerable[_A_], ys: IEnumerable[_B_]) -> None:
    def arrow_172(unit_var0: None, x: _A_, y: _B_, action: Callable[[_A_, _B_], None]=action, xs: IEnumerable[_A_]=xs, ys: IEnumerable[_B_]=ys) -> None:
        action(x, y)
    
    fold2(arrow_172, None, xs, ys)


def iterate_indexed(action: Callable[[int, _A_], None], xs: IEnumerable[_A_]) -> None:
    def arrow_173(i: int, x: _A_, action: Callable[[int, _A_], None]=action, xs: IEnumerable[_A_]=xs) -> int:
        action(i, x)
        return i + 1
    
    ignore(fold(arrow_173, 0, xs))


def iterate_indexed2(action: Callable[[int, _A_, _B_], None], xs: IEnumerable[_A_], ys: IEnumerable[_B_]) -> None:
    def arrow_174(i: int, x: _A_, y: _B_, action: Callable[[int, _A_, _B_], None]=action, xs: IEnumerable[_A_]=xs, ys: IEnumerable[_B_]=ys) -> int:
        action(i, x, y)
        return i + 1
    
    ignore(fold2(arrow_174, 0, xs, ys))


def try_last(xs: IEnumerable[_T]) -> Optional[_T]:
    with of_seq(xs) as e:
        def loop(acc_mut: Optional[_T]=None, xs: IEnumerable[_T]=xs) -> _T:
            while True:
                (acc,) = (acc_mut,)
                if not e.System_Collections_IEnumerator_MoveNext():
                    return acc
                
                else: 
                    acc_mut = e.System_Collections_Generic_IEnumerator_00601_get_Current()
                    continue
                
                break
        
        loop : Callable[[_T], _T] = loop
        if e.System_Collections_IEnumerator_MoveNext():
            return some(loop(e.System_Collections_Generic_IEnumerator_00601_get_Current()))
        
        else: 
            return None
        


def last(xs: IEnumerable[_T]) -> _T:
    match_value : Optional[_T] = try_last(xs)
    if match_value is None:
        raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "source")
    
    else: 
        return value_1(match_value)
    


def length(xs: IEnumerable[_T]) -> int:
    if is_array_like(xs):
        return len(xs)
    
    elif isinstance(xs, FSharpList):
        return length_1(xs)
    
    else: 
        with of_seq(xs) as e:
            count : int = 0
            while e.System_Collections_IEnumerator_MoveNext():
                count = (count + 1) or 0
            return count
    


def map(mapping: Callable[[_T], _U], xs: IEnumerable[_T]) -> IEnumerable[_U]:
    def arrow_175(mapping: Callable[[_T], _U]=mapping, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_176(e: IEnumerator[_T], mapping: Callable[[_T], _U]=mapping, xs: IEnumerable[_T]=xs) -> Optional[_U]:
        return some(mapping(e.System_Collections_Generic_IEnumerator_00601_get_Current())) if e.System_Collections_IEnumerator_MoveNext() else None
    
    def arrow_177(e_1: IEnumerator[_T], mapping: Callable[[_T], _U]=mapping, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate(arrow_175, arrow_176, arrow_177)


def map_indexed(mapping: Callable[[int, _T], _U], xs: IEnumerable[_T]) -> IEnumerable[_U]:
    def arrow_178(mapping: Callable[[int, _T], _U]=mapping, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_179(i: int, e: IEnumerator[_T], mapping: Callable[[int, _T], _U]=mapping, xs: IEnumerable[_T]=xs) -> Optional[_U]:
        return some(mapping(i, e.System_Collections_Generic_IEnumerator_00601_get_Current())) if e.System_Collections_IEnumerator_MoveNext() else None
    
    def arrow_180(e_1: IEnumerator[_T], mapping: Callable[[int, _T], _U]=mapping, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate_indexed(arrow_178, arrow_179, arrow_180)


def indexed(xs: IEnumerable[_T]) -> IEnumerable[Tuple[int, _T]]:
    def mapping(i: int, x: _T, xs: IEnumerable[_T]=xs) -> Tuple[int, _T]:
        return (i, x)
    
    return map_indexed(mapping, xs)


def map2(mapping: Callable[[_T1, _T2], _U], xs: IEnumerable[_T1], ys: IEnumerable[_T2]) -> IEnumerable[_U]:
    def arrow_181(mapping: Callable[[_T1, _T2], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys) -> Tuple[IEnumerator[_T1], IEnumerator[_T2]]:
        return (of_seq(xs), of_seq(ys))
    
    def arrow_182(tupled_arg: Tuple[IEnumerator[_T1], IEnumerator[_T2]], mapping: Callable[[_T1, _T2], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys) -> Optional[_U]:
        e1 : IEnumerator[_T1] = tupled_arg[0]
        e2 : IEnumerator[_T2] = tupled_arg[1]
        return some(mapping(e1.System_Collections_Generic_IEnumerator_00601_get_Current(), e2.System_Collections_Generic_IEnumerator_00601_get_Current())) if (e2.System_Collections_IEnumerator_MoveNext() if e1.System_Collections_IEnumerator_MoveNext() else False) else None
    
    def arrow_183(tupled_arg_1: Tuple[IEnumerator[_T1], IEnumerator[_T2]], mapping: Callable[[_T1, _T2], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys) -> None:
        try: 
            dispose_2(tupled_arg_1[0])
        
        finally: 
            dispose_2(tupled_arg_1[1])
        
    
    return generate(arrow_181, arrow_182, arrow_183)


def map_indexed2(mapping: Callable[[int, _T1, _T2], _U], xs: IEnumerable[_T1], ys: IEnumerable[_T2]) -> IEnumerable[_U]:
    def arrow_184(mapping: Callable[[int, _T1, _T2], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys) -> Tuple[IEnumerator[_T1], IEnumerator[_T2]]:
        return (of_seq(xs), of_seq(ys))
    
    def arrow_185(i: int, tupled_arg: Tuple[IEnumerator[_T1], IEnumerator[_T2]], mapping: Callable[[int, _T1, _T2], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys) -> Optional[_U]:
        e1 : IEnumerator[_T1] = tupled_arg[0]
        e2 : IEnumerator[_T2] = tupled_arg[1]
        return some(mapping(i, e1.System_Collections_Generic_IEnumerator_00601_get_Current(), e2.System_Collections_Generic_IEnumerator_00601_get_Current())) if (e2.System_Collections_IEnumerator_MoveNext() if e1.System_Collections_IEnumerator_MoveNext() else False) else None
    
    def arrow_186(tupled_arg_1: Tuple[IEnumerator[_T1], IEnumerator[_T2]], mapping: Callable[[int, _T1, _T2], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys) -> None:
        try: 
            dispose_2(tupled_arg_1[0])
        
        finally: 
            dispose_2(tupled_arg_1[1])
        
    
    return generate_indexed(arrow_184, arrow_185, arrow_186)


def map3(mapping: Callable[[_T1, _T2, _T3], _U], xs: IEnumerable[_T1], ys: IEnumerable[_T2], zs: IEnumerable[_T3]) -> IEnumerable[_U]:
    def arrow_187(mapping: Callable[[_T1, _T2, _T3], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys, zs: IEnumerable[_T3]=zs) -> Tuple[IEnumerator[_T1], IEnumerator[_T2], IEnumerator[_T3]]:
        return (of_seq(xs), of_seq(ys), of_seq(zs))
    
    def arrow_188(tupled_arg: Tuple[IEnumerator[_T1], IEnumerator[_T2], IEnumerator[_T3]], mapping: Callable[[_T1, _T2, _T3], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys, zs: IEnumerable[_T3]=zs) -> Optional[_U]:
        e1 : IEnumerator[_T1] = tupled_arg[0]
        e2 : IEnumerator[_T2] = tupled_arg[1]
        e3 : IEnumerator[_T3] = tupled_arg[2]
        return some(mapping(e1.System_Collections_Generic_IEnumerator_00601_get_Current(), e2.System_Collections_Generic_IEnumerator_00601_get_Current(), e3.System_Collections_Generic_IEnumerator_00601_get_Current())) if (e3.System_Collections_IEnumerator_MoveNext() if (e2.System_Collections_IEnumerator_MoveNext() if e1.System_Collections_IEnumerator_MoveNext() else False) else False) else None
    
    def arrow_189(tupled_arg_1: Tuple[IEnumerator[_T1], IEnumerator[_T2], IEnumerator[_T3]], mapping: Callable[[_T1, _T2, _T3], _U]=mapping, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys, zs: IEnumerable[_T3]=zs) -> None:
        try: 
            dispose_2(tupled_arg_1[0])
        
        finally: 
            try: 
                dispose_2(tupled_arg_1[1])
            
            finally: 
                dispose_2(tupled_arg_1[2])
            
        
    
    return generate(arrow_187, arrow_188, arrow_189)


def read_only(xs: IEnumerable[_T]) -> IEnumerable[_T]:
    check_non_null("source", xs)
    def arrow_190(x: Optional[_T]=None, xs: IEnumerable[_T]=xs) -> _T:
        return x
    
    return map(arrow_190, xs)


def expr_191(gen0: TypeInfo) -> TypeInfo:
    return class_type("SeqModule.CachedSeq`1", [gen0], CachedSeq_1)


class CachedSeq_1(IDisposable, Generic[_T]):
    def __init__(self, cleanup: Callable[[], None], res: IEnumerable[_T]) -> None:
        self.cleanup = cleanup
        self.res = res
    
    def Dispose(self) -> None:
        _ : CachedSeq_1[_T] = self
        _.cleanup()
    
    def GetEnumerator(self) -> IEnumerator[_T]:
        _ : CachedSeq_1[_T] = self
        return get_enumerator(_.res)
    
    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())
    
    def System_Collections_IEnumerable_GetEnumerator(self) -> IEnumerator[Any]:
        _ : CachedSeq_1[_T] = self
        return get_enumerator(_.res)
    

CachedSeq_1_reflection = expr_191

def CachedSeq_1__ctor_Z7A8347D4(cleanup: Callable[[], None], res: IEnumerable[_T]) -> CachedSeq_1[_T]:
    return CachedSeq_1(cleanup, res)


def CachedSeq_1__Clear(_: CachedSeq_1[_T]) -> None:
    _.cleanup()


def cache(source: IEnumerable[_T]) -> IEnumerable[_T]:
    check_non_null("source", source)
    prefix : List[_T] = []
    enumerator_r : Optional[Optional[IEnumerator[_T]]] = None
    def cleanup(source: IEnumerable[_T]=source) -> None:
        def arrow_192(__unit: Any=None) -> None:
            nonlocal enumerator_r
            clear(prefix)
            (pattern_matching_result, e) = (None, None)
            if enumerator_r is not None:
                if value_1(enumerator_r) is not None:
                    pattern_matching_result = 0
                    e = value_1(enumerator_r)
                
                else: 
                    pattern_matching_result = 1
                
            
            else: 
                pattern_matching_result = 1
            
            if pattern_matching_result == 0:
                dispose_2(e)
            
            enumerator_r = None
        
        Operators_Lock(prefix, arrow_192)
    
    def arrow_194(i_1: int, source: IEnumerable[_T]=source) -> Optional[Tuple[_T, int]]:
        def arrow_193(__unit: Any=None) -> Optional[Tuple[_T, int]]:
            nonlocal enumerator_r
            if i_1 < len(prefix):
                return (prefix[i_1], i_1 + 1)
            
            else: 
                if i_1 >= len(prefix):
                    opt_enumerator_2 : Optional[IEnumerator[_T]]
                    if enumerator_r is not None:
                        opt_enumerator_2 = value_1(enumerator_r)
                    
                    else: 
                        opt_enumerator : Optional[IEnumerator[_T]] = get_enumerator(source)
                        enumerator_r = some(opt_enumerator)
                        opt_enumerator_2 = opt_enumerator
                    
                    if opt_enumerator_2 is None:
                        pass
                    
                    else: 
                        enumerator : IEnumerator[_T] = opt_enumerator_2
                        if enumerator.System_Collections_IEnumerator_MoveNext():
                            (prefix.append(enumerator.System_Collections_Generic_IEnumerator_00601_get_Current()))
                        
                        else: 
                            dispose_2(enumerator)
                            enumerator_r = some(None)
                        
                    
                
                return ((prefix[i_1], i_1 + 1)) if (i_1 < len(prefix)) else None
            
        
        return Operators_Lock(prefix, arrow_193)
    
    return CachedSeq_1__ctor_Z7A8347D4(cleanup, unfold(arrow_194, 0))


def all_pairs(xs: IEnumerable[_T1], ys: IEnumerable[_T2]) -> IEnumerable[Tuple[_T1, _T2]]:
    ys_cache : IEnumerable[_T2] = cache(ys)
    def arrow_195(xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys) -> IEnumerable[Tuple[_T1, _T2]]:
        def mapping_1(x: Optional[_A_]=None) -> IEnumerable[Tuple[_A_, _T2]]:
            def mapping(y: Optional[_T2]=None, x: _A_=x) -> Tuple[_A_, _T2]:
                return (x, y)
            
            return map(mapping, ys_cache)
        
        return concat(map(mapping_1, xs))
    
    return delay(arrow_195)


def map_fold(mapping: Callable[[_STATE, _T], Tuple[_RESULT, _STATE]], state: _STATE, xs: IEnumerable[_T]) -> Tuple[IEnumerable[_RESULT], _STATE]:
    pattern_input : Tuple[List[_RESULT], _STATE] = map_fold_1(mapping, state, to_array(xs), None)
    return (read_only(pattern_input[0]), pattern_input[1])


def map_fold_back(mapping: Callable[[_T, _STATE], Tuple[_RESULT, _STATE]], xs: IEnumerable[_T], state: _STATE) -> Tuple[IEnumerable[_RESULT], _STATE]:
    pattern_input : Tuple[List[_RESULT], _STATE] = map_fold_back_1(mapping, to_array(xs), state, None)
    return (read_only(pattern_input[0]), pattern_input[1])


def try_pick(chooser: Callable[[_T], Optional[_A_]], xs: IEnumerable[_T]) -> Optional[_A_]:
    with of_seq(xs) as e:
        res : Optional[_A_] = None
        while e.System_Collections_IEnumerator_MoveNext() if (res is None) else False:
            res = chooser(e.System_Collections_Generic_IEnumerator_00601_get_Current())
        return res


def pick(chooser: Callable[[_T], Optional[_A_]], xs: IEnumerable[_T]) -> _A_:
    match_value : Optional[_A_] = try_pick(chooser, xs)
    if match_value is None:
        return index_not_found()
    
    else: 
        return value_1(match_value)
    


def reduce(folder: Callable[[_T, _T], _T], xs: IEnumerable[_T]) -> _T:
    with of_seq(xs) as e:
        def loop(acc_mut: Optional[_T]=None, folder: Callable[[_T, _T], _T]=folder, xs: IEnumerable[_T]=xs) -> _T:
            while True:
                (acc,) = (acc_mut,)
                if e.System_Collections_IEnumerator_MoveNext():
                    acc_mut = folder(acc, e.System_Collections_Generic_IEnumerator_00601_get_Current())
                    continue
                
                else: 
                    return acc
                
                break
        
        loop : Callable[[_T], _T] = loop
        if e.System_Collections_IEnumerator_MoveNext():
            return loop(e.System_Collections_Generic_IEnumerator_00601_get_Current())
        
        else: 
            raise Exception(SR_inputSequenceEmpty)
        


def reduce_back(folder: Callable[[_T, _T], _T], xs: IEnumerable[_T]) -> _T:
    arr : List[_T] = to_array(xs)
    if len(arr) > 0:
        return reduce_back_1(folder, arr)
    
    else: 
        raise Exception(SR_inputSequenceEmpty)
    


def replicate(n: int, x: _A_) -> IEnumerable[_A_]:
    def arrow_196(_arg1: int, n: int=n, x: _A_=x) -> _A_:
        return x
    
    return initialize(n, arrow_196)


def reverse(xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_197(xs: IEnumerable[_T]=xs) -> IEnumerable[_T]:
        return of_array(reverse_1(to_array(xs)))
    
    return delay(arrow_197)


def scan(folder: Callable[[_STATE, _T], _STATE], state: _STATE, xs: IEnumerable[_T]) -> IEnumerable[_STATE]:
    def arrow_198(folder: Callable[[_STATE, _T], _STATE]=folder, state: _STATE=state, xs: IEnumerable[_T]=xs) -> IEnumerable[_STATE]:
        acc : _STATE = state
        def mapping(x: Optional[_T]=None) -> _STATE:
            nonlocal acc
            acc = folder(acc, x)
            return acc
        
        return concat([singleton(state), map(mapping, xs)])
    
    return delay(arrow_198)


def scan_back(folder: Callable[[_T, _STATE], _STATE], xs: IEnumerable[_T], state: _STATE) -> IEnumerable[_STATE]:
    def arrow_199(folder: Callable[[_T, _STATE], _STATE]=folder, xs: IEnumerable[_T]=xs, state: _STATE=state) -> IEnumerable[_STATE]:
        return of_array(scan_back_1(folder, to_array(xs), state, None))
    
    return delay(arrow_199)


def skip(count: int, source: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_200(count: int=count, source: IEnumerable[_T]=source) -> IEnumerator[_T]:
        e : IEnumerator[_T] = of_seq(source)
        try: 
            for _ in range(1, count + 1, 1):
                if not e.System_Collections_IEnumerator_MoveNext():
                    raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "source")
                
            def compensation(__unit: Any=None) -> None:
                pass
            
            return Enumerator_enumerateThenFinally(compensation, e)
        
        except Exception as match_value:
            dispose_2(e)
            raise match_value
        
    
    return mk_seq(arrow_200)


def skip_while(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_201(predicate: Callable[[_T], bool]=predicate, xs: IEnumerable[_T]=xs) -> IEnumerable[_T]:
        skipped : bool = True
        def f(x: Optional[_T]=None) -> bool:
            nonlocal skipped
            if skipped:
                skipped = predicate(x)
            
            return not skipped
        
        return filter(f, xs)
    
    return delay(arrow_201)


def tail(xs: IEnumerable[_T]) -> IEnumerable[_T]:
    return skip(1, xs)


def take(count: int, xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_202(count: int=count, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_203(i: int, e: IEnumerator[_T], count: int=count, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        if i < count:
            if e.System_Collections_IEnumerator_MoveNext():
                return some(e.System_Collections_Generic_IEnumerator_00601_get_Current())
            
            else: 
                raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "source")
            
        
        else: 
            return None
        
    
    def arrow_204(e_1: IEnumerator[_T], count: int=count, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate_indexed(arrow_202, arrow_203, arrow_204)


def take_while(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_205(predicate: Callable[[_T], bool]=predicate, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_206(e: IEnumerator[_T], predicate: Callable[[_T], bool]=predicate, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        return some(e.System_Collections_Generic_IEnumerator_00601_get_Current()) if (predicate(e.System_Collections_Generic_IEnumerator_00601_get_Current()) if e.System_Collections_IEnumerator_MoveNext() else False) else None
    
    def arrow_207(e_1: IEnumerator[_T], predicate: Callable[[_T], bool]=predicate, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate(arrow_205, arrow_206, arrow_207)


def truncate(count: int, xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_208(count: int=count, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_209(i: int, e: IEnumerator[_T], count: int=count, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        return some(e.System_Collections_Generic_IEnumerator_00601_get_Current()) if (e.System_Collections_IEnumerator_MoveNext() if (i < count) else False) else None
    
    def arrow_210(e_1: IEnumerator[_T], count: int=count, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate_indexed(arrow_208, arrow_209, arrow_210)


def zip(xs: IEnumerable[_T1], ys: IEnumerable[_T2]) -> IEnumerable[Tuple[_T1, _T2]]:
    def arrow_211(x: _T1, y: _T2, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys) -> Tuple[_T1, _T2]:
        return (x, y)
    
    return map2(arrow_211, xs, ys)


def zip3(xs: IEnumerable[_T1], ys: IEnumerable[_T2], zs: IEnumerable[_T3]) -> IEnumerable[Tuple[_T1, _T2, _T3]]:
    def arrow_212(x: _T1, y: _T2, z: _T3, xs: IEnumerable[_T1]=xs, ys: IEnumerable[_T2]=ys, zs: IEnumerable[_T3]=zs) -> Tuple[_T1, _T2, _T3]:
        return (x, y, z)
    
    return map3(arrow_212, xs, ys, zs)


def collect(mapping: Callable[[_T], IEnumerable[_U]], xs: IEnumerable[_T]) -> IEnumerable[_U]:
    def arrow_213(mapping: Callable[[_T], IEnumerable[_U]]=mapping, xs: IEnumerable[_T]=xs) -> IEnumerable[_U]:
        return concat(map(mapping, xs))
    
    return delay(arrow_213)


def where(predicate: Callable[[_T], bool], xs: IEnumerable[_T]) -> IEnumerable[_T]:
    return filter(predicate, xs)


def pairwise(xs: IEnumerable[_T]) -> IEnumerable[Tuple[_T, _T]]:
    def arrow_214(xs: IEnumerable[_T]=xs) -> IEnumerable[Tuple[_T, _T]]:
        return of_array(pairwise_1(to_array(xs)))
    
    return delay(arrow_214)


def split_into(chunks: int, xs: IEnumerable[_T]) -> IEnumerable[IEnumerable[_T]]:
    def arrow_215(chunks: int=chunks, xs: IEnumerable[_T]=xs) -> IEnumerable[IEnumerable[_T]]:
        def mapping(arr: List[_T]) -> IEnumerable[_T]:
            return of_array(arr)
        
        return of_array(map_1(mapping, split_into_1(chunks, to_array(xs)), None))
    
    return delay(arrow_215)


def windowed(window_size: int, xs: IEnumerable[_T]) -> IEnumerable[IEnumerable[_T]]:
    def arrow_216(window_size: int=window_size, xs: IEnumerable[_T]=xs) -> IEnumerable[IEnumerable[_T]]:
        def mapping(arr: List[_T]) -> IEnumerable[_T]:
            return of_array(arr)
        
        return of_array(map_1(mapping, windowed_1(window_size, to_array(xs)), None))
    
    return delay(arrow_216)


def transpose(xss: IEnumerable[_A_]) -> IEnumerable[IEnumerable[_T]]:
    def arrow_217(xss: IEnumerable[_A_]=xss) -> IEnumerable[IEnumerable[_T]]:
        def mapping_1(arr: List[_T]) -> IEnumerable[_T]:
            return of_array(arr)
        
        def mapping(xs_1: Optional[_A_]=None) -> List[_T]:
            return to_array(xs_1)
        
        return of_array(map_1(mapping_1, transpose_1(map_1(mapping, to_array(xss), None), None), None))
    
    return delay(arrow_217)


def sort_with(comparer: Callable[[_T, _T], int], xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_218(comparer: Callable[[_T, _T], int]=comparer, xs: IEnumerable[_T]=xs) -> IEnumerable[_T]:
        arr : List[_T] = to_array(xs)
        arr.sort()
        return of_array(arr)
    
    return delay(arrow_218)


def sort(xs: IEnumerable[_T], comparer: IComparer[_T]) -> IEnumerable[_T]:
    def arrow_219(x: _T, y: _T, xs: IEnumerable[_T]=xs, comparer: IComparer[_T]=comparer) -> int:
        return comparer.Compare(x, y)
    
    return sort_with(arrow_219, xs)


def sort_by(projection: Callable[[_T], _U], xs: IEnumerable[_T], comparer: IComparer[_U]) -> IEnumerable[_T]:
    def arrow_220(x: _T, y: _T, projection: Callable[[_T], _U]=projection, xs: IEnumerable[_T]=xs, comparer: IComparer[_U]=comparer) -> int:
        return comparer.Compare(projection(x), projection(y))
    
    return sort_with(arrow_220, xs)


def sort_descending(xs: IEnumerable[_T], comparer: IComparer[_T]) -> IEnumerable[_T]:
    def arrow_221(x: _T, y: _T, xs: IEnumerable[_T]=xs, comparer: IComparer[_T]=comparer) -> int:
        return comparer.Compare(x, y) * -1
    
    return sort_with(arrow_221, xs)


def sort_by_descending(projection: Callable[[_T], _U], xs: IEnumerable[_T], comparer: IComparer[_U]) -> IEnumerable[_T]:
    def arrow_222(x: _T, y: _T, projection: Callable[[_T], _U]=projection, xs: IEnumerable[_T]=xs, comparer: IComparer[_U]=comparer) -> int:
        return comparer.Compare(projection(x), projection(y)) * -1
    
    return sort_with(arrow_222, xs)


def sum(xs: IEnumerable[_T], adder: IGenericAdder_1[_T]) -> _T:
    def arrow_223(acc: _T, x: _T, xs: IEnumerable[_T]=xs, adder: IGenericAdder_1[_T]=adder) -> _T:
        return adder.Add(acc, x)
    
    return fold(arrow_223, adder.GetZero(), xs)


def sum_by(f: Callable[[_T], _U], xs: IEnumerable[_T], adder: IGenericAdder_1[_U]) -> _U:
    def arrow_224(acc: _U, x: _T, f: Callable[[_T], _U]=f, xs: IEnumerable[_T]=xs, adder: IGenericAdder_1[_U]=adder) -> _U:
        return adder.Add(acc, f(x))
    
    return fold(arrow_224, adder.GetZero(), xs)


def max_by(projection: Callable[[_T], _U], xs: IEnumerable[_T], comparer: IComparer[_U]) -> _T:
    def arrow_225(x: _T, y: _T, projection: Callable[[_T], _U]=projection, xs: IEnumerable[_T]=xs, comparer: IComparer[_U]=comparer) -> _T:
        return y if (comparer.Compare(projection(y), projection(x)) > 0) else x
    
    return reduce(arrow_225, xs)


def max(xs: IEnumerable[_T], comparer: IComparer[_T]) -> _T:
    def arrow_226(x: _T, y: _T, xs: IEnumerable[_T]=xs, comparer: IComparer[_T]=comparer) -> _T:
        return y if (comparer.Compare(y, x) > 0) else x
    
    return reduce(arrow_226, xs)


def min_by(projection: Callable[[_T], _U], xs: IEnumerable[_T], comparer: IComparer[_U]) -> _T:
    def arrow_227(x: _T, y: _T, projection: Callable[[_T], _U]=projection, xs: IEnumerable[_T]=xs, comparer: IComparer[_U]=comparer) -> _T:
        return x if (comparer.Compare(projection(y), projection(x)) > 0) else y
    
    return reduce(arrow_227, xs)


def min(xs: IEnumerable[_T], comparer: IComparer[_T]) -> _T:
    def arrow_228(x: _T, y: _T, xs: IEnumerable[_T]=xs, comparer: IComparer[_T]=comparer) -> _T:
        return x if (comparer.Compare(y, x) > 0) else y
    
    return reduce(arrow_228, xs)


def average(xs: IEnumerable[_T], averager: IGenericAverager_1[_T]) -> _T:
    count : int = 0
    def folder(acc: _T, x: _T, xs: IEnumerable[_T]=xs, averager: IGenericAverager_1[_T]=averager) -> _T:
        nonlocal count
        count = (count + 1) or 0
        return averager.Add(acc, x)
    
    total : _T = fold(folder, averager.GetZero(), xs)
    if count == 0:
        raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "source")
    
    else: 
        return averager.DivideByInt(total, count)
    


def average_by(f: Callable[[_T], _U], xs: IEnumerable[_T], averager: IGenericAverager_1[_U]) -> _U:
    count : int = 0
    def arrow_229(acc: _U, x: _T, f: Callable[[_T], _U]=f, xs: IEnumerable[_T]=xs, averager: IGenericAverager_1[_U]=averager) -> _U:
        nonlocal count
        count = (count + 1) or 0
        return averager.Add(acc, f(x))
    
    total : _U = fold(arrow_229, averager.GetZero(), xs)
    if count == 0:
        raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "source")
    
    else: 
        return averager.DivideByInt(total, count)
    


def permute(f: Callable[[int], int], xs: IEnumerable[_T]) -> IEnumerable[_T]:
    def arrow_230(f: Callable[[int], int]=f, xs: IEnumerable[_T]=xs) -> IEnumerable[_T]:
        return of_array(permute_1(f, to_array(xs)))
    
    return delay(arrow_230)


def chunk_by_size(chunk_size: int, xs: IEnumerable[_T]) -> IEnumerable[IEnumerable[_T]]:
    def arrow_231(chunk_size: int=chunk_size, xs: IEnumerable[_T]=xs) -> IEnumerable[IEnumerable[_T]]:
        def mapping(arr: List[_T]) -> IEnumerable[_T]:
            return of_array(arr)
        
        return of_array(map_1(mapping, chunk_by_size_1(chunk_size, to_array(xs)), None))
    
    return delay(arrow_231)


def insert_at(index: int, y: _T, xs: IEnumerable[_T]) -> IEnumerable[_T]:
    is_done : bool = False
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    def arrow_232(index: int=index, y: _T=y, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_233(i: int, e: IEnumerator[_T], index: int=index, y: _T=y, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        nonlocal is_done
        if e.System_Collections_IEnumerator_MoveNext() if (True if is_done else (i < index)) else False:
            return some(e.System_Collections_Generic_IEnumerator_00601_get_Current())
        
        elif i == index:
            is_done = True
            return some(y)
        
        else: 
            if not is_done:
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
            
            return None
        
    
    def arrow_234(e_1: IEnumerator[_T], index: int=index, y: _T=y, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate_indexed(arrow_232, arrow_233, arrow_234)


def insert_many_at(index: int, ys: IEnumerable[_T], xs: IEnumerable[_T]) -> IEnumerable[_T]:
    status : int = -1
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    def arrow_235(index: int=index, ys: IEnumerable[_T]=ys, xs: IEnumerable[_T]=xs) -> Tuple[IEnumerator[_T], IEnumerator[_T]]:
        return (of_seq(xs), of_seq(ys))
    
    def arrow_236(i: int, tupled_arg: Tuple[IEnumerator[_T], IEnumerator[_T]], index: int=index, ys: IEnumerable[_T]=ys, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        nonlocal status
        e1 : IEnumerator[_T] = tupled_arg[0]
        e2 : IEnumerator[_T] = tupled_arg[1]
        if i == index:
            status = 0
        
        inserted : Optional[_T]
        if status == 0:
            if e2.System_Collections_IEnumerator_MoveNext():
                inserted = some(e2.System_Collections_Generic_IEnumerator_00601_get_Current())
            
            else: 
                status = 1
                inserted = None
            
        
        else: 
            inserted = None
        
        if inserted is None:
            if e1.System_Collections_IEnumerator_MoveNext():
                return some(e1.System_Collections_Generic_IEnumerator_00601_get_Current())
            
            else: 
                if status < 1:
                    raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
                
                return None
            
        
        else: 
            return some(value_1(inserted))
        
    
    def arrow_237(tupled_arg_1: Tuple[IEnumerator[_T], IEnumerator[_T]], index: int=index, ys: IEnumerable[_T]=ys, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(tupled_arg_1[0])
        dispose_2(tupled_arg_1[1])
    
    return generate_indexed(arrow_235, arrow_236, arrow_237)


def remove_at(index: int, xs: IEnumerable[_T]) -> IEnumerable[_T]:
    is_done : bool = False
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    def arrow_238(index: int=index, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_239(i: int, e: IEnumerator[_T], index: int=index, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        nonlocal is_done
        if e.System_Collections_IEnumerator_MoveNext() if (True if is_done else (i < index)) else False:
            return some(e.System_Collections_Generic_IEnumerator_00601_get_Current())
        
        elif e.System_Collections_IEnumerator_MoveNext() if (i == index) else False:
            is_done = True
            return some(e.System_Collections_Generic_IEnumerator_00601_get_Current()) if e.System_Collections_IEnumerator_MoveNext() else None
        
        else: 
            if not is_done:
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
            
            return None
        
    
    def arrow_240(e_1: IEnumerator[_T], index: int=index, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate_indexed(arrow_238, arrow_239, arrow_240)


def remove_many_at(index: int, count: int, xs: IEnumerable[_T]) -> IEnumerable[_T]:
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    def arrow_241(index: int=index, count: int=count, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_242(i: int, e: IEnumerator[_T], index: int=index, count: int=count, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        if i < index:
            if e.System_Collections_IEnumerator_MoveNext():
                return some(e.System_Collections_Generic_IEnumerator_00601_get_Current())
            
            else: 
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
            
        
        else: 
            if i == index:
                for _ in range(1, count + 1, 1):
                    if not e.System_Collections_IEnumerator_MoveNext():
                        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "count")
                    
            
            return some(e.System_Collections_Generic_IEnumerator_00601_get_Current()) if e.System_Collections_IEnumerator_MoveNext() else None
        
    
    def arrow_243(e_1: IEnumerator[_T], index: int=index, count: int=count, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate_indexed(arrow_241, arrow_242, arrow_243)


def update_at(index: int, y: _T, xs: IEnumerable[_T]) -> IEnumerable[_T]:
    is_done : bool = False
    if index < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    def arrow_244(index: int=index, y: _T=y, xs: IEnumerable[_T]=xs) -> IEnumerator[_T]:
        return of_seq(xs)
    
    def arrow_245(i: int, e: IEnumerator[_T], index: int=index, y: _T=y, xs: IEnumerable[_T]=xs) -> Optional[_T]:
        nonlocal is_done
        if e.System_Collections_IEnumerator_MoveNext() if (True if is_done else (i < index)) else False:
            return some(e.System_Collections_Generic_IEnumerator_00601_get_Current())
        
        elif e.System_Collections_IEnumerator_MoveNext() if (i == index) else False:
            is_done = True
            return some(y)
        
        else: 
            if not is_done:
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
            
            return None
        
    
    def arrow_246(e_1: IEnumerator[_T], index: int=index, y: _T=y, xs: IEnumerable[_T]=xs) -> None:
        dispose_2(e_1)
    
    return generate_indexed(arrow_244, arrow_245, arrow_246)





from abc import abstractmethod
from typing import (TypeVar, Protocol, Generic, Any)

_T = TypeVar("_T")

class IGenericAdder_1(Protocol, Generic[_T]):
    @abstractmethod
    def Add(self, __arg0: _T, __arg1: _T) -> _T:
        ...
    
    @abstractmethod
    def GetZero(self) -> _T:
        ...
    

class IGenericAverager_1(Protocol, Generic[_T]):
    @abstractmethod
    def Add(self, __arg0: _T, __arg1: _T) -> _T:
        ...
    
    @abstractmethod
    def DivideByInt(self, __arg0: _T, __arg1: int) -> _T:
        ...
    
    @abstractmethod
    def GetZero(self) -> _T:
        ...
    

class Symbol_wellknown(Protocol):
    @property
    @abstractmethod
    def Symbol_toStringTag(self) -> str:
        ...
    

class IJsonSerializable(Protocol):
    @abstractmethod
    def to_json(self, key: str) -> Any:
        ...
    

SR_indexOutOfBounds : str = "The index was outside the range of elements in the collection."

SR_inputWasEmpty : str = "Collection was empty."

SR_inputMustBeNonNegative : str = "The input must be non-negative."

SR_inputSequenceEmpty : str = "The input sequence was empty."

SR_inputSequenceTooLong : str = "The input sequence contains more than one element."

SR_keyNotFoundAlt : str = "An index satisfying the predicate was not found in the collection."

SR_differentLengths : str = "The collections had different lengths."

SR_notEnoughElements : str = "The input sequence has an insufficient number of elements."




import unicodedata
from enum import Enum
from typing import Dict, Union


class UnicodeCategory(Enum):
    """Defines the Unicode category of a character.

    https://docs.microsoft.com/en-us/dotnet/api/system.globalization.unicodecategory?view=net-6.0
    """

    UppercaseLetter = 0
    LowercaseLetter = 1
    TitlecaseLetter = 2
    ModifierLetter = 3
    OtherLetter = 4
    NonSpacingMark = 5
    SpacingCombiningMark = 6
    EnclosingMark = 7
    DecimalDigitNumber = 8
    LetterNumber = 9
    OtherNumber = 10
    SpaceSeparator = 11
    LineSeparator = 12
    ParagraphSeparator = 13
    Control = 14
    Format = 15
    Surrogate = 16
    PrivateUse = 17
    ConnectorPunctuation = 18
    DashPunctuation = 19
    OpenPunctuation = 20
    ClosePunctuation = 21
    InitialQuotePunctuation = 22
    FinalQuotePunctuation = 23
    OtherPunctuation = 24
    MathSymbol = 25
    CurrencySymbol = 26
    ModifierSymbol = 27
    OtherSymbol = 28
    OtherNotAssigned = 29


unicodeCategory2Python: Dict[str, UnicodeCategory] = {
    "Ll": UnicodeCategory.LowercaseLetter,
    "Lu": UnicodeCategory.UppercaseLetter,
    "Nd": UnicodeCategory.DecimalDigitNumber,
    "Cc": UnicodeCategory.Control,
    "Pd": UnicodeCategory.DashPunctuation,
    "Sc": UnicodeCategory.CurrencySymbol,
    "Pc": UnicodeCategory.ConnectorPunctuation,
    "Cf": UnicodeCategory.Format,
    "Pi": UnicodeCategory.InitialQuotePunctuation,
    "Po": UnicodeCategory.OtherPunctuation,
    "Nl": UnicodeCategory.LetterNumber,
    "Zl": UnicodeCategory.LineSeparator,
    "Zs": UnicodeCategory.SpaceSeparator,
    "Sm": UnicodeCategory.MathSymbol,
    "Sk": UnicodeCategory.ModifierSymbol,
    "Mn": UnicodeCategory.NonSpacingMark,
    "Lo": UnicodeCategory.OtherLetter,
    "No": UnicodeCategory.OtherLetter,
}


def char_code_at(s: str, index: int):
    if index >= 0 and index < len(s):
        return ord(s[index : index + 1])
    else:
        raise ValueError("Index out of range.")


def get_unicode_category(s: str, index: int = 0):
    ret = unicodeCategory2Python.get(unicodedata.category(s[index : index + 1]))
    if ret:
        return ret.value
    raise ValueError("Fable error")


IS_LETTER_MASK = (
    0
    | 1 << UnicodeCategory.UppercaseLetter.value
    | 1 << UnicodeCategory.LowercaseLetter.value
    | 1 << UnicodeCategory.TitlecaseLetter.value
    | 1 << UnicodeCategory.ModifierLetter.value
    | 1 << UnicodeCategory.OtherLetter.value
)
IS_PUNCTUATION_MASK = (
    0
    | 1 << UnicodeCategory.ConnectorPunctuation.value
    | 1 << UnicodeCategory.DashPunctuation.value
    | 1 << UnicodeCategory.OpenPunctuation.value
    | 1 << UnicodeCategory.ClosePunctuation.value
    | 1 << UnicodeCategory.InitialQuotePunctuation.value
    | 1 << UnicodeCategory.FinalQuotePunctuation.value
    | 1 << UnicodeCategory.OtherPunctuation.value
)

IS_SEPARATOR_MASK = (
    0
    | 1 << UnicodeCategory.SpaceSeparator.value
    | 1 << UnicodeCategory.LineSeparator.value
    | 1 << UnicodeCategory.ParagraphSeparator.value
)
IS_NUMBER_MASK = (
    0
    | 1 << UnicodeCategory.DecimalDigitNumber.value
    | 1 << UnicodeCategory.LetterNumber.value
    | 1 << UnicodeCategory.OtherNumber.value
)
IS_SYMBOL_MASK = (
    0
    | 1 << UnicodeCategory.MathSymbol.value
    | 1 << UnicodeCategory.CurrencySymbol.value
    | 1 << UnicodeCategory.ModifierSymbol.value
    | 1 << UnicodeCategory.OtherSymbol.value
)
IS_WHITESPACE_MASK = (
    0
    | 1 << UnicodeCategory.SpaceSeparator.value
    | 1 << UnicodeCategory.LineSeparator.value
    | 1 << UnicodeCategory.ParagraphSeparator.value
)

IS_DIGIT_MASK = 1 << UnicodeCategory.DecimalDigitNumber.value
IS_CONTROL_MASK = 1 << UnicodeCategory.Control.value
IS_LETTER_OR_DIGIT_MASK = IS_LETTER_MASK | IS_DIGIT_MASK
IS_UPPER_MASK = 1 << UnicodeCategory.UppercaseLetter.value
IS_LOWER_MASK = 1 << UnicodeCategory.LowercaseLetter.value


def is_digit(s: str, index: int = 0) -> bool:
    return s[index : index + 1].isdigit()


def is_letter(s: str, index: int = 0) -> bool:
    return s[index : index + 1].isalpha()


def is_letter_or_digit(s: str, index: int = 0) -> bool:
    test = 1 << get_unicode_category(s, index)
    return bool(test & IS_LETTER_OR_DIGIT_MASK)


def is_control(s: str, index: int = 0):
    test = 1 << get_unicode_category(s, index)
    return bool(test & IS_CONTROL_MASK)


def is_upper(s: str, index: int = 0) -> bool:
    test = 1 << get_unicode_category(s, index)
    return bool(test & IS_UPPER_MASK)


def is_lower(s: str, index: int = 0) -> bool:
    test = 1 << get_unicode_category(s, index)
    return bool(test & IS_LOWER_MASK)


def is_punctuation(s: str, index: int = 0) -> bool:
    test = 1 << get_unicode_category(s, index)
    return bool(test & IS_PUNCTUATION_MASK)


def is_separator(s: str, index: int = 0) -> bool:
    test = 1 << get_unicode_category(s, index)
    return bool(test & IS_SEPARATOR_MASK)


def is_number(s: str, index: int = 0) -> bool:
    test = 1 << get_unicode_category(s, index)
    return bool(test & IS_NUMBER_MASK)


def is_symbol(s: str, index: int = 0) -> bool:
    test = 1 << get_unicode_category(s, index)
    return bool(test & IS_SYMBOL_MASK)


def is_white_space(s: str, index: int = 0) -> bool:
    test = 1 << get_unicode_category(s, index)
    if test & IS_WHITESPACE_MASK:
        return True

    cp = char_code_at(s, index)
    return (0x09 <= cp <= 0x0D) or cp == 0x85 or cp == 0xA0


def is_high_surrogate(s: str, index: int = 0) -> bool:
    cp = char_code_at(s, index)
    return 0xD800 <= cp <= 0xDBFF


def is_low_surrogate(s: str, index: int = 0) -> bool:
    cp = char_code_at(s, index)
    return 0xDC00 <= cp <= 0xDFFF


def is_surrogate(s: str, index: int = 0) -> bool:
    cp = char_code_at(s, index)
    return 0xD800 <= cp <= 0xDFFF


def is_surrogate_pair(s: str, index: Union[str, int]) -> bool:
    if isinstance(index, int):
        return is_high_surrogate(s, index) and is_low_surrogate(s, index + 1)

    return is_high_surrogate(s) and is_low_surrogate(index)


def parse(input: str) -> str:
    if len(input) == 1:
        return input[0]
    else:
        raise ValueError("String must be exactly one character long.")


__all__ = [
    "char_code_at",
    "get_unicode_category",
    "is_control",
    "is_letter_or_digit",
    "is_letter",
    "is_digit",
    "is_upper",
    "is_lower",
    "is_separator",
    "is_punctuation",
    "is_number",
    "is_symbol",
    "is_separator",
    "is_white_space",
    "is_surrogate",
    "is_low_surrogate",
    "is_high_surrogate",
    "is_surrogate_pair",
    "parse",
]



from __future__ import annotations
from typing import (Optional, TypeVar, List, Tuple, Any)
from ..fable_library.array import fold as fold_1
from ..fable_library.option import (to_array, value as value_2)
from ..fable_library.seq import fold
from ..fable_library.string import (replicate, to_text, printf)
from ..fable_library.util import int32_to_string
from ..fs_hafas_python.parse.journey import distance_of_journey
from .types_hafas_client import (Location as Location_1, Stop as Stop_1, Station as Station_1, Alternative as Alternative_1, Hint, Status, Warning as Warning_1, Leg as Leg_1, Line as Line_1, StopOver as StopOver_1, Trip as Trip_1, Journey as Journey_1, Price, Journeys as Journeys_1, Duration as Duration_1, Movement as Movement_1)

_A = TypeVar("_A")

nl : str = "\n"

def printf_s(ident: int, prefix: str, s: Optional[str]=None) -> str:
    ident_s : str = replicate(ident, " ")
    def folder(s_1: str, value: str, ident: int=ident, prefix: str=prefix, s: Optional[str]=s) -> str:
        return s_1 + (to_text(printf("%s%s%s%s"))(ident_s)(prefix)(value)(")") if (prefix == "(") else to_text(printf("%s%s%s"))(ident_s)(prefix)(value))
    
    return fold(folder, "", to_array(s))


def printfn_s(ident: int, prefix: str, s: Optional[str]=None) -> str:
    return printf_s(ident, prefix, s) + nl


def printfn_arr_l(ident: int, prefix: str, arr: Optional[List[_A]]=None) -> str:
    ident_s : str = replicate(ident, " ")
    (pattern_matching_result, value_1) = (None, None)
    if arr is not None:
        if len(arr) > 0:
            pattern_matching_result = 0
            value_1 = arr
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        def arrow_525(ident: int=ident, prefix: str=prefix, arr: Optional[List[_A]]=arr) -> str:
            arg30 : int = len(value_1) or 0
            return to_text(printf("%s%s%d"))(ident_s)(prefix)(arg30)
        
        return arrow_525() + nl
    
    elif pattern_matching_result == 1:
        return ""
    


def printfn_b(ident: int, prefix: str, b: Optional[bool]=None) -> str:
    ident_s : str = replicate(ident, " ")
    (pattern_matching_result, value_1) = (None, None)
    if b is not None:
        if b:
            pattern_matching_result = 0
            value_1 = b
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        return to_text(printf("%s%s%b"))(ident_s)(prefix)(value_1) + nl
    
    elif pattern_matching_result == 1:
        return ""
    


def print_distance(ident: int, d: Optional[int]=None) -> str:
    ident_s : str = replicate(ident, " ")
    (pattern_matching_result, d_2) = (None, None)
    if d is not None:
        if d > 0:
            pattern_matching_result = 0
            d_2 = d
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        km : float = d_2 / 1000
        return to_text(printf("%s%s%0.3f"))(ident_s)("distance: ")(km) + nl
    
    elif pattern_matching_result == 1:
        return ""
    


def print_lon_lat(ident: int, lon: Optional[float]=None, lat: Optional[float]=None) -> str:
    ident_s : str = replicate(ident, " ")
    match_value : Tuple[Optional[float], Optional[float]] = (lon, lat)
    (pattern_matching_result, lat_1, lon_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            lat_1 = match_value[1]
            lon_1 = match_value[0]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        return to_text(printf("%s%s%f,%f"))(ident_s)("lonlat: ")(lon_1)(lat_1)
    
    elif pattern_matching_result == 1:
        return ""
    


def print_name_id(ident: int, name: Optional[str]=None, id: Optional[str]=None) -> str:
    return printf_s(ident, "", name) + printf_s(ident, "(", id)


def Location(ident: int, location: Location_1) -> str:
    def arrow_526(ident: int=ident, location: Location_1=location) -> str:
        match_value : Tuple[Optional[str], Optional[str]] = (location.name, location.address)
        return (print_name_id(ident, location.name, location.id) + " ") if (match_value[0] is not None) else ((print_name_id(ident, location.address, location.id) + " ") if (match_value[1] is not None) else "")
    
    return ((arrow_526() + print_lon_lat(ident, location.longitude, location.latitude)) + nl) + print_distance(ident + 2, location.distance)


def Stop(ident: int, stop: Stop_1) -> str:
    def arrow_527(ident: int=ident, stop: Stop_1=stop) -> str:
        match_value : Optional[Location_1] = stop.location
        if match_value is None:
            return ""
        
        else: 
            location : Location_1 = match_value
            return " " + print_lon_lat(ident, location.longitude, location.latitude)
        
    
    return ((print_name_id(ident, stop.name, stop.id) + arrow_527()) + nl) + print_distance(ident + 2, stop.distance)


def Station(ident: int, station: Station_1) -> str:
    return (print_name_id(ident, station.name, station.id) + nl) + print_distance(ident + 2, station.distance)


def Alternative(ident: int, alternative: Alternative_1) -> str:
    def arrow_528(ident: int=ident, alternative: Alternative_1=alternative) -> str:
        match_value : Optional[Any] = alternative.stop
        if match_value is not None:
            stop : Any = value_2(match_value)
            return (printf_s(ident, "stop: ", "") + Stop(0, stop)) if isinstance(stop, Stop_1) else ""
        
        else: 
            return ""
        
    
    return (printfn_s(ident, "direction: ", alternative.direction) + printfn_s(ident, "when: ", alternative.when)) + arrow_528()


def Comment(ident: int, s: str) -> str:
    if len(s) > 0:
        return printfn_s(ident, "---", "")
    
    else: 
        return ""
    


def Remark(ident: int, remark: Any=None) -> str:
    if isinstance(remark, Hint):
        return printfn_s(ident, "hint: ", remark.text)
    
    elif isinstance(remark, Status):
        return printfn_s(ident, "status: ", remark.text)
    
    else: 
        return ""
    


def Warning(ident: int, w: Warning_1) -> str:
    return (printfn_s(ident, "warning: ", w.summary) + printfn_s(ident + 2, "validFrom: ", w.valid_from)) + printfn_s(ident + 2, "validUntil: ", w.valid_until)


def Remarks(ident: int, remarks: Optional[List[Any]]=None) -> str:
    if remarks is None:
        return ""
    
    else: 
        def folder(s: str, r: Any=None, ident: int=ident, remarks: Optional[List[Any]]=remarks) -> str:
            return s + Remark(ident + 2, r)
        
        return fold_1(folder, "", remarks)
    


def ProductOfLeg(ident: int, leg: Leg_1) -> str:
    match_value : Optional[Line_1] = leg.line
    if match_value is None:
        return ""
    
    else: 
        match_value_1 : Optional[str] = match_value.product
        if match_value_1 is None:
            return ""
        
        else: 
            return printfn_s(ident, "product: ", match_value_1)
        
    


def StopOverStop(ident: int, so: StopOver_1) -> str:
    match_value : Optional[Any] = so.stop
    if match_value is not None:
        if isinstance(value_2(match_value), Station_1):
            s_1 : Station_1 = value_2(match_value)
            return printf_s(ident, "", "") + Station(0, s_1)
        
        else: 
            s : Stop_1 = value_2(match_value)
            return printf_s(ident, "", "") + Stop(0, s)
        
    
    else: 
        return ""
    


def StopOver(ident: int, so: StopOver_1) -> str:
    def arrow_531(ident: int=ident, so: StopOver_1=so) -> str:
        match_value : Optional[Any] = so.stop
        def arrow_529(__unit: Any=None) -> str:
            s_1 : Station_1 = value_2(match_value)
            return printf_s(ident, "origin: ", "") + Station(0, s_1)
        
        def arrow_530(__unit: Any=None) -> str:
            s : Stop_1 = value_2(match_value)
            return printf_s(ident, "origin: ", "") + Stop(0, s)
        
        return (arrow_529() if isinstance(value_2(match_value), Station_1) else arrow_530()) if (match_value is not None) else ""
    
    def arrow_532(ident: int=ident, so: StopOver_1=so) -> str:
        match_value_1 : Optional[str] = so.departure
        if match_value_1 is None:
            return printfn_s(ident, "arrival: ", so.arrival)
        
        else: 
            departure : str = match_value_1
            return printfn_s(ident, "departure: ", so.departure)
        
    
    return arrow_531() + arrow_532()


def StopOvers(ident: int, stop_overs: Optional[List[StopOver_1]]=None) -> str:
    if stop_overs is None:
        return ""
    
    else: 
        stop_overs_1 : List[StopOver_1] = stop_overs
        def folder(s: str, l: StopOver_1, ident: int=ident, stop_overs: Optional[List[StopOver_1]]=stop_overs) -> str:
            return (s + Comment(ident + 2, s)) + StopOver(ident + 2, l)
        
        return printfn_s(ident, "stopOvers: ", "") + fold_1(folder, "", stop_overs_1)
    


def Leg(ident: int, leg: Leg_1, short: bool) -> str:
    def arrow_536(ident: int=ident, leg: Leg_1=leg, short: bool=short) -> str:
        match_value : Optional[Any] = leg.origin
        def arrow_533(__unit: Any=None) -> str:
            s : Stop_1 = value_2(match_value)
            return printf_s(ident, "origin: ", "") + Stop(0, s)
        
        def arrow_534(__unit: Any=None) -> str:
            s_1 : Station_1 = value_2(match_value)
            return printf_s(ident, "origin: ", "") + Station(0, s_1)
        
        def arrow_535(__unit: Any=None) -> str:
            l : Location_1 = value_2(match_value)
            return printf_s(ident, "origin: ", "") + Location(0, l)
        
        return (arrow_533() if isinstance(value_2(match_value), Stop_1) else (arrow_534() if isinstance(value_2(match_value), Station_1) else arrow_535())) if (match_value is not None) else ""
    
    def arrow_537(ident: int=ident, leg: Leg_1=leg, short: bool=short) -> str:
        match_value_1 : Optional[Any] = leg.destination
        (pattern_matching_result, s_2) = (None, None)
        if match_value_1 is not None:
            if isinstance(value_2(match_value_1), Stop_1):
                pattern_matching_result = 0
                s_2 = value_2(match_value_1)
            
            else: 
                pattern_matching_result = 1
            
        
        else: 
            pattern_matching_result = 1
        
        if pattern_matching_result == 0:
            return printf_s(ident, "destination: ", "") + Stop(0, s_2)
        
        elif pattern_matching_result == 1:
            return ""
        
    
    def arrow_538(ident: int=ident, leg: Leg_1=leg, short: bool=short) -> str:
        match_value_2 : Optional[Location_1] = leg.current_location
        if match_value_2 is None:
            return ""
        
        else: 
            location : Location_1 = match_value_2
            return ((replicate(ident, " ") + "currentLocation: ") + print_lon_lat(0, location.longitude, location.latitude)) + nl
        
    
    def arrow_539(ident: int=ident, leg: Leg_1=leg, short: bool=short) -> str:
        match_value_3 : Optional[Line_1] = leg.line
        (pattern_matching_result_1, line_1) = (None, None)
        if match_value_3 is not None:
            if match_value_3.name is not None:
                pattern_matching_result_1 = 0
                line_1 = match_value_3
            
            else: 
                pattern_matching_result_1 = 1
            
        
        else: 
            pattern_matching_result_1 = 1
        
        if pattern_matching_result_1 == 0:
            return printfn_s(ident, "Line: ", line_1.name)
        
        elif pattern_matching_result_1 == 1:
            return ""
        
    
    return (((((((printfn_s(ident, "tripId: ", leg.trip_id) + arrow_536()) + arrow_537()) + printfn_s(ident, "departure: ", leg.departure)) + printfn_s(ident, "arrival: ", leg.arrival)) + ("" if short else StopOvers(ident, leg.stopovers))) + printfn_b(ident, "cancelled: ", leg.cancelled)) + arrow_538()) + ((ProductOfLeg(ident, leg) + printfn_s(ident, "loadFactor: ", leg.load_factor)) if short else (((((arrow_539() + printfn_b(ident, "walking: ", leg.walking)) + printfn_b(ident, "transfer: ", leg.transfer)) + ProductOfLeg(ident, leg)) + printfn_s(ident, "loadFactor: ", leg.load_factor)) + Remarks(ident, leg.remarks)))


def Trip(ident: int, trip: Trip_1) -> str:
    def arrow_540(ident: int=ident, trip: Trip_1=trip) -> str:
        match_value : Optional[Any] = trip.origin
        (pattern_matching_result, s, s_1) = (None, None, None)
        if match_value is not None:
            if isinstance(value_2(match_value), Stop_1):
                pattern_matching_result = 0
                s = value_2(match_value)
            
            elif isinstance(value_2(match_value), Station_1):
                pattern_matching_result = 1
                s_1 = value_2(match_value)
            
            else: 
                pattern_matching_result = 2
            
        
        else: 
            pattern_matching_result = 2
        
        if pattern_matching_result == 0:
            return printf_s(ident, "origin: ", "") + Stop(0, s)
        
        elif pattern_matching_result == 1:
            return printf_s(ident, "origin: ", "") + Station(0, s_1)
        
        elif pattern_matching_result == 2:
            return ""
        
    
    def arrow_541(ident: int=ident, trip: Trip_1=trip) -> str:
        match_value_1 : Optional[Any] = trip.destination
        (pattern_matching_result_1, s_2) = (None, None)
        if match_value_1 is not None:
            if isinstance(value_2(match_value_1), Stop_1):
                pattern_matching_result_1 = 0
                s_2 = value_2(match_value_1)
            
            else: 
                pattern_matching_result_1 = 1
            
        
        else: 
            pattern_matching_result_1 = 1
        
        if pattern_matching_result_1 == 0:
            return printf_s(ident, "destination: ", "") + Stop(0, s_2)
        
        elif pattern_matching_result_1 == 1:
            return ""
        
    
    def arrow_542(ident: int=ident, trip: Trip_1=trip) -> str:
        match_value_2 : Optional[Line_1] = trip.line
        (pattern_matching_result_2, line_1) = (None, None)
        if match_value_2 is not None:
            if match_value_2.name is not None:
                pattern_matching_result_2 = 0
                line_1 = match_value_2
            
            else: 
                pattern_matching_result_2 = 1
            
        
        else: 
            pattern_matching_result_2 = 1
        
        if pattern_matching_result_2 == 0:
            return printfn_s(ident, "Line: ", line_1.name)
        
        elif pattern_matching_result_2 == 1:
            return ""
        
    
    return ((((((((arrow_540() + arrow_541()) + printfn_s(ident, "departure: ", trip.departure)) + printfn_s(ident, "arrival: ", trip.arrival)) + printfn_arr_l(ident, "stopovers: ", trip.stopovers)) + arrow_542()) + printfn_b(ident, "cancelled: ", trip.cancelled)) + printfn_b(ident, "walking: ", trip.walking)) + printfn_b(ident, "transfer: ", trip.transfer)) + Remarks(ident, trip.remarks)


def Legs(ident: int, legs: List[Leg_1], short: bool) -> str:
    def folder(s: str, l: Leg_1, ident: int=ident, legs: List[Leg_1]=legs, short: bool=short) -> str:
        return (s + Comment(ident + 2, s)) + Leg(ident + 2, l, short)
    
    return fold_1(folder, "", legs)


def JourneyLegs(ident: int, journey: Journey_1) -> str:
    return printfn_s(ident, "jouney:", "") + Legs(ident, journey.legs, True)


def Journey(ident: int, journey: Journey_1) -> str:
    short : bool = True
    price_1 : str
    match_value : Optional[Price] = journey.price
    if match_value is None:
        price_1 = ""
    
    else: 
        price : Price = match_value
        ident_s_1 : str = replicate(ident + 2, " ")
        price_1 = to_text(printf("%sprice: %.2f %s"))(ident_s_1)(price.amount)(price.currency) + nl
    
    def arrow_543(ident: int=ident, journey: Journey_1=journey) -> str:
        distance : float = distance_of_journey(journey)
        if distance > 0:
            ident_s : str = replicate(ident + 2, " ")
            return to_text(printf("%sdistance: %.2f"))(ident_s)(distance) + nl
        
        else: 
            return ""
        
    
    def arrow_544(ident: int=ident, journey: Journey_1=journey) -> str:
        match_value_1 : Tuple[bool, Optional[str]] = (short, journey.refresh_token)
        (pattern_matching_result,) = (None,)
        if match_value_1[0]:
            pattern_matching_result = 1
        
        elif match_value_1[1] is not None:
            pattern_matching_result = 0
        
        else: 
            pattern_matching_result = 1
        
        if pattern_matching_result == 0:
            return printfn_s(ident + 2, "refreshToken: \u0027", match_value_1[1] + "\u0027")
        
        elif pattern_matching_result == 1:
            return ""
        
    
    return (((printfn_s(ident, "jouney:", "") + Legs(ident, journey.legs, short)) + price_1) + arrow_543()) + arrow_544()


def JourneyItems(ident: int, journeys: List[Journey_1]) -> str:
    def folder(s: str, j: Journey_1, ident: int=ident, journeys: List[Journey_1]=journeys) -> str:
        return s + Journey(ident, j)
    
    return fold_1(folder, "", journeys)


def Journeys(journeys: Journeys_1) -> str:
    def folder(s: str, value: List[Journey_1], journeys: Journeys_1=journeys) -> str:
        return s + JourneyItems(0, value)
    
    return fold(folder, "", to_array(journeys.journeys))


def U2StationStop(ident: int, location: Optional[Any]=None) -> str:
    if location is not None:
        if isinstance(value_2(location), Stop_1):
            s_1 : Stop_1 = value_2(location)
            return Stop(ident + 2, s_1)
        
        else: 
            s : Station_1 = value_2(location)
            return Station(ident + 2, s)
        
    
    else: 
        return ""
    


def U3StationStopLocation(ident: int, location: Any=None) -> str:
    if isinstance(location, Location_1):
        return Location(ident + 2, location)
    
    elif isinstance(location, Stop_1):
        return Stop(ident + 2, location)
    
    else: 
        return ""
    


def Duration(ident: int, duration: Duration_1) -> str:
    def folder(s: str, j: Any=None, ident: int=ident, duration: Duration_1=duration) -> str:
        return s + U3StationStopLocation(ident + 2, j)
    
    return printfn_s(ident, "duration: ", int32_to_string(duration.duration)) + fold_1(folder, "", duration.stations)


def Directions(ident: int, directions: Optional[List[str]]=None) -> str:
    if directions is None:
        return ""
    
    else: 
        def folder(s: str, j: str, ident: int=ident, directions: Optional[List[str]]=directions) -> str:
            return s + printfn_s(ident + 2, "", j)
        
        return fold_1(folder, "", directions)
    


def Line(ident: int, l: Line_1) -> str:
    return (printfn_s(ident, "name: ", l.name) + printfn_arr_l(ident, "directions: ", l.directions)) + Directions(ident + 2, l.directions)


def Movement(ident: int, m: Movement_1, with_stopovers: bool) -> str:
    def arrow_545(ident: int=ident, m: Movement_1=m, with_stopovers: bool=with_stopovers) -> str:
        match_value : Optional[Line_1] = m.line
        (pattern_matching_result, line_1) = (None, None)
        if match_value is not None:
            if match_value.name is not None:
                pattern_matching_result = 0
                line_1 = match_value
            
            else: 
                pattern_matching_result = 1
            
        
        else: 
            pattern_matching_result = 1
        
        if pattern_matching_result == 0:
            return printfn_s(ident, "Line: ", line_1.name)
        
        elif pattern_matching_result == 1:
            return ""
        
    
    def arrow_546(ident: int=ident, m: Movement_1=m, with_stopovers: bool=with_stopovers) -> str:
        match_value_1 : Optional[Location_1] = m.location
        if match_value_1 is None:
            return ""
        
        else: 
            location : Location_1 = match_value_1
            return print_lon_lat(ident, location.longitude, location.latitude)
        
    
    return (((printfn_s(ident, "tripId: ", m.trip_id) + printfn_s(ident, "direction: ", m.direction)) + arrow_545()) + arrow_546()) + ((printfn_arr_l(ident, "stopovers: ", m.next_stopovers) + StopOvers(ident, m.next_stopovers)) if with_stopovers else "")


def Locations(locations: List[Any]) -> str:
    def folder(s: str, j: Any=None, locations: List[Any]=locations) -> str:
        return s + U3StationStopLocation(0, j)
    
    return fold_1(folder, "", locations)


def Durations(durations: List[Duration_1]) -> str:
    def folder(s: str, j: Duration_1, durations: List[Duration_1]=durations) -> str:
        return s + Duration(0, j)
    
    return fold_1(folder, "", durations)


def MovementsWithStopovers(durations: List[Movement_1]) -> str:
    def folder(s: str, j: Movement_1, durations: List[Movement_1]=durations) -> str:
        return s + Movement(0, j, True)
    
    return fold_1(folder, "", durations)


def Movements(durations: List[Movement_1]) -> str:
    def folder(s: str, j: Movement_1, durations: List[Movement_1]=durations) -> str:
        return s + Movement(0, j, False)
    
    return fold_1(folder, "", durations)


def Alternatives(alternatives: List[Alternative_1]) -> str:
    def folder(s: str, a: Alternative_1, alternatives: List[Alternative_1]=alternatives) -> str:
        return s + Alternative(0, a)
    
    return fold_1(folder, "", alternatives)


def Trips(trips: List[Trip_1]) -> str:
    def folder(s: str, t: Trip_1, trips: List[Trip_1]=trips) -> str:
        return s + Trip(0, t)
    
    return fold_1(folder, "", trips)


def Warnings(warnings: List[Warning_1]) -> str:
    def folder(s: str, t: Warning_1, warnings: List[Warning_1]=warnings) -> str:
        return s + Warning(0, t)
    
    return fold_1(folder, "", warnings)


def Lines(lines: List[Line_1]) -> str:
    def folder(s: str, t: Line_1, lines: List[Line_1]=lines) -> str:
        return s + Line(0, t)
    
    return fold_1(folder, "", lines)





from typing import TYPE_CHECKING, Any, Dict, Set, TypeVar

from .types import FSharpRef

_K = TypeVar("_K")
_V = TypeVar("_V")

if TYPE_CHECKING:

    class FSharpMap(Dict[_K, _V]):
        ...


else:
    from .map import FSharpMap


def add_to_set(v: _V, st: Set[_V]) -> bool:
    if v in st:
        return False

    st.add(v)
    return True


def add_to_dict(di: Dict[_K, _V], k: _K, v: _V) -> None:
    if k in di:
        raise Exception(
            "An item with the same key has already been added. Key: " + str(k)
        )

    di[k] = v


def try_get_value(
    map: FSharpMap[_K, _V], key: _K, default_value: FSharpRef[_V]
) -> bool:
    print([map], key)
    if key in map.keys():

        default_value.contents = map[key]
        return True

    return False


def get_item_from_dict(map: Dict[_K, _V], key: _K) -> _V:
    if key in map:
        return map[key]
    else:
        raise Exception(f"The given key '{key}' was not present in the dictionary.")


def contains_value(v: _V, map: Dict[Any, _V]) -> bool:
    return v in map.values()



from __future__ import annotations
from typing import (Optional, List, MutableSequence, Callable)
from ...fable_library.array import map as map_1
from ...fable_library.option import map
from ...fable_library.string import replace
from ...fable_library.types import Int32Array
from ...fs_hafas_python.context import (Context, Profile__get_parseDateTime, Profile__get_parseBitmask)
from ...fs_hafas_python.extra_types import IndexMap_2
from ...fs_hafas_python.lib.transformations import Default_Warning
from ...fs_hafas_python.types_hafas_client import (IcoCrd, Edge, Event, Line, Warning)
from ...fs_hafas_python.types_raw_hafas_client import (RawHimMsgEdge, RawHimMsgEvent, RawHim, RawHimMsgCat, RawCommon)
from .common import (get_element_at_some, map_index_array, to_option)

def parse_msg_edge(ctx: Context, e: RawHimMsgEdge) -> Edge:
    ico_crd : Optional[IcoCrd] = IcoCrd(e.ico_crd.x, e.ico_crd.y, None)
    return Edge(get_element_at_some(e.f_loc_x, ctx.common.locations), get_element_at_some(e.t_loc_x, ctx.common.locations), get_element_at_some(e.ico_x, ctx.common.icons), e.dir, ico_crd)


def parse_msg_event(ctx: Context, e: RawHimMsgEvent) -> Event:
    return Event(get_element_at_some(e.f_loc_x, ctx.common.locations), get_element_at_some(e.t_loc_x, ctx.common.locations), Profile__get_parseDateTime(ctx.profile)(ctx)(e.f_date)(e.f_time)(None), Profile__get_parseDateTime(ctx.profile)(ctx)(e.t_date)(e.t_time)(None), [])


def parse_date_time(ctx: Context, date: Optional[str]=None, time: Optional[str]=None) -> Optional[str]:
    if date is not None:
        date_1 : str = date
        return Profile__get_parseDateTime(ctx.profile)(ctx)(date_1)(time)(None)
    
    else: 
        return None
    


def parse_warning(ctx: Context, w: RawHim) -> Warning:
    def mapping(p_cls: int, ctx: Context=ctx, w: RawHim=w) -> IndexMap_2[str, bool]:
        return Profile__get_parseBitmask(ctx.profile)(ctx)(p_cls)
    
    products : Optional[IndexMap_2[str, bool]] = map(mapping, w.prod)
    def mapping_1(c_1: RawHimMsgCat, ctx: Context=ctx, w: RawHim=w) -> int:
        return c_1.id
    
    def get_target_array(c: RawCommon, ctx: Context=ctx, w: RawHim=w) -> Optional[List[RawHimMsgCat]]:
        return c.him_msg_cat_l
    
    categories : MutableSequence[int] = map_1(mapping_1, map_index_array(ctx.res.common, get_target_array, w.cat_ref_l), Int32Array)
    def mapping_2(e: RawHimMsgEvent, ctx: Context=ctx, w: RawHim=w) -> Event:
        return parse_msg_event(ctx, e)
    
    def get_target_array_1(c_2: RawCommon, ctx: Context=ctx, w: RawHim=w) -> Optional[List[RawHimMsgEvent]]:
        return c_2.him_msg_event_l
    
    events : Optional[List[Event]] = to_option(map_1(mapping_2, map_index_array(ctx.res.common, get_target_array_1, w.event_ref_l), None))
    def mapping_3(e_1: RawHimMsgEdge, ctx: Context=ctx, w: RawHim=w) -> Edge:
        return parse_msg_edge(ctx, e_1)
    
    def get_target_array_2(c_3: RawCommon, ctx: Context=ctx, w: RawHim=w) -> Optional[List[RawHimMsgEdge]]:
        return c_3.him_msg_edge_l
    
    edges : Optional[List[Edge]] = to_option(map_1(mapping_3, map_index_array(ctx.res.common, get_target_array_2, w.edge_ref_l), None))
    def get_target_array_3(_arg1: RawCommon, ctx: Context=ctx, w: RawHim=w) -> Optional[List[Line]]:
        return ctx.common.lines
    
    affected_lines : Optional[List[Line]] = to_option(map_index_array(ctx.res.common, get_target_array_3, w.aff_prod_ref_l))
    def br_to_newline(s: Optional[str]=None, ctx: Context=ctx, w: RawHim=w) -> Optional[str]:
        def mapping_4(s_1: str, s: Optional[str]=s) -> str:
            return replace(s_1, "\u003cbr\u003e", "\n")
        
        return map(mapping_4, s)
    
    return Warning(Default_Warning.type, w.hid, get_element_at_some(w.ico_x, ctx.common.icons), br_to_newline(w.head), br_to_newline(w.text), Default_Warning.category, w.prio, products, edges, events, parse_date_time(ctx, w.s_date, w.s_time), parse_date_time(ctx, w.e_date, w.e_time), parse_date_time(ctx, w.l_mod_date, w.l_mod_time), w.comp, categories, affected_lines, Default_Warning.from_stops, Default_Warning.to_stops)


